
#include <linux/config.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/in.h>
#include <linux/rtnetlink.h>
#include <linux/notifier.h>

#include <asm/uaccess.h> /* for copy_from_user */

#ifdef ACTION_TEC_VIRTUAL_ETH
#include <linux/cpmac_veth.h>

/* Port numbers start from 1 */
struct net_device *cpmac_veth_devices[CPMAC_VETH_DEV_NUM];

static int cpmac_veth_device_event(struct notifier_block *, 
                                   unsigned long, 
                                   void *);

struct notifier_block cpmac_veth_notifier_block = {
	notifier_call: cpmac_veth_device_event,
};

struct net_device *find_cpmac_veth_dev(struct net_device *real_dev,
                                       unsigned int port_num)
{
    return cpmac_veth_devices[port_num-1];
}

/*  
 * Create a virtual network device attached to the real 
 * CPMAC network device.
 * Returns the device that was created, or NULL if there was
 * an error of some kind.
 */
struct net_device *register_cpmac_veth_device(const char *eth_if_name, 
                                              unsigned int port_num)
{
	struct net_device *new_dev;
	struct net_device *real_dev; /* the ethernet device */
	int malloc_size = 0;

#ifdef CPMAC_VETH_DEBUG
	printk(KERN_DEBUG "register_cpmac_veth_device: if_name -:%s:-	port: %i\n",
	       eth_if_name, port_num);
#endif /* CPMAC_VETH_DEBUG */

	if (port_num > CPMAC_VETH_DEV_NUM)
		goto out_ret_null;

	/* find the device relating to eth_if_name. */
	real_dev = dev_get_by_name(eth_if_name);
	if (!real_dev)
		goto out_ret_null;
#ifdef CPMAC_VETH_DEBUG
    printk(KERN_DEBUG "register_cpmac_veth_device: real_dev=%s/%d\n", real_dev->name, real_dev->ifindex);
#endif /* CPMAC_VETH_DEBUG */

	if (find_cpmac_veth_dev(real_dev, port_num))
    {
#ifdef CPMAC_VETH_DEBUG
		/* Already registered. */
		printk(KERN_DEBUG "register_cpmac_veth_device: Already had VETH registered\n");
#endif /* CPMAC_VETH_DEBUG */
		dev_put(real_dev);
		return NULL;
	}

	malloc_size = sizeof(struct net_device);
	new_dev = (struct net_device *) kmalloc(malloc_size, GFP_KERNEL);
	if (new_dev == NULL)
		goto out_put_dev;
	memset(new_dev, 0, malloc_size);

    malloc_size = sizeof(struct cpmac_veth_dev_info);
	new_dev->priv = (struct cpmac_veth_dev_info *) kmalloc(malloc_size, 
                                                           GFP_KERNEL);
	if (new_dev->priv == NULL)
    {
        kfree(new_dev);
		goto out_put_dev;
    }
	memset(new_dev->priv, 0, malloc_size);

	/* 
     * set us up to not use a Qdisc, as the underlying Hardware device
	 * can do all the queueing we could want.
	 */
	/* new_dev->qdisc_sleeping = &noqueue_qdisc;   Not needed it seems. */
	new_dev->tx_queue_len = 0; /* This should effectively give us no queue. */

    /* Set up the fields for the device. */

    /* name will look like: cp1, cp2 ... */
    sprintf(new_dev->name, "cp%i", port_num);
#ifdef CPMAC_VETH_DEBUG
	printk(KERN_DEBUG "register_cpmac_veth_device: Allocated new name -:%s:-\n", new_dev->name);
#endif /* CPMAC_VETH_DEBUG */

	/* set up method calls */
	new_dev->init = cpmac_veth_dev_init;
	new_dev->destructor = cpmac_veth_dev_destruct;
	new_dev->features |= NETIF_F_DYNALLOC ; 
	    
	new_dev->get_stats = cpmac_veth_dev_get_stats;
	    
	new_dev->flags = real_dev->flags;
	new_dev->flags &= ~IFF_UP;

	new_dev->priv_flags |= IFF_VETH;
				
	new_dev->mtu = real_dev->mtu;
	new_dev->change_mtu = cpmac_veth_dev_change_mtu;

	new_dev->type = real_dev->type;

	new_dev->hard_header_len = real_dev->hard_header_len;
	new_dev->hard_header = real_dev->hard_header;
	new_dev->rebuild_header = real_dev->rebuild_header;
    new_dev->hard_start_xmit = cpmac_veth_dev_hard_start_xmit;
	new_dev->hard_header_parse = real_dev->hard_header_parse;

	memcpy(new_dev->broadcast, real_dev->broadcast, real_dev->addr_len);
	memcpy(new_dev->dev_addr, real_dev->dev_addr, real_dev->addr_len);
	new_dev->addr_len = real_dev->addr_len;

	new_dev->open = cpmac_veth_dev_open;
	new_dev->stop = cpmac_veth_dev_stop;

	new_dev->set_mac_address = cpmac_veth_dev_set_mac_address;
	new_dev->set_multicast_list = cpmac_veth_dev_set_multicast_list;

	CPMAC_VETH_DEV_INFO(new_dev)->port_num = port_num;
	CPMAC_VETH_DEV_INFO(new_dev)->real_dev = real_dev;
	CPMAC_VETH_DEV_INFO(new_dev)->flags = 0; /* unused for now */

	cpmac_veth_devices[port_num-1] = new_dev;

	/* create it's proc entry - Not needed as of now */

    /* 
     * Register the netdevice for this virtual ethernet interface
     * with the Linux Network Stack
     */
	rtnl_lock();
	register_netdevice(new_dev);
	rtnl_unlock();
	    
#ifdef CPMAC_VETH_DEBUG
	printk(KERN_DEBUG "register_cpmac_veth_device: Allocated new device successfully - %s/%d/%d    port=%d \n", new_dev->name, new_dev->ifindex, new_dev->hard_header_len, CPMAC_VETH_DEV_INFO(new_dev)->port_num);
#endif /* CPMAC_VETH_DEBUG */

	return new_dev;

out_put_dev:
	dev_put(real_dev);

out_ret_null:
	return NULL;
}

int unregister_cpmac_veth_dev(int real_dev_ifindex, 
                              unsigned int port_num, 
                              int do_dev_put, 
                              int do_locks)
{
	struct net_device *dev = NULL;

#ifdef CPMAC_VETH_DEBUG
	printk(KERN_DEBUG "unregister_cpmac_veth_dev: port_num: %i\n", port_num);
#endif /* CPMAC_VETH_DEBUG */

    /* sanity check */
    if (port_num > CPMAC_VETH_DEV_NUM)
        return -EINVAL;

    dev = cpmac_veth_devices[port_num-1];
    if (dev) 
    {
        cpmac_veth_devices[port_num-1] = NULL;

        /* 
         * Take it out of the global list of devices.
         * NOTE:  This deletes dev, don't access it again!!
         */
        if (do_dev_put)
            dev_put(dev);

        if (do_locks) 
        {
            rtnl_lock();
            unregister_netdevice(dev);
            rtnl_unlock();
        } 
        else 
        {
            unregister_netdevice(dev);
        }
    }
#ifdef CPMAC_VETH_DEBUG
	printk(KERN_DEBUG "unregister_cpmac_veth_dev: Unregistered successfully\n");
#endif /* CPMAC_VETH_DEBUG */
    
    return 0;
}

int unregister_cpmac_veth_device(const char *veth_if_name)
{
    unsigned int port_num = 0;
	struct net_device *dev = NULL, *vdev = NULL;

#ifdef CPMAC_VETH_DEBUG
	printk(KERN_DEBUG "unregister_cpmac_veth_device: -:%s:-\n", veth_if_name);
#endif /* CPMAC_VETH_DEBUG */

	dev = dev_get_by_name(veth_if_name);
	if (dev) 
    {
		if (dev->priv_flags & IFF_VETH) 
        {
			return unregister_cpmac_veth_dev(
                    CPMAC_VETH_DEV_INFO(dev)->real_dev->ifindex,
                    (unsigned short)(CPMAC_VETH_DEV_INFO(dev)->port_num),
                    1 /* do dev_put */, 1 /* do locking */);
		} 
        else 
        {
			printk(KERN_ERR "unregister_cpmac_veth_device: Tried to remove a non-existent device, name: %s\n", dev->name);
			dev_put(dev);
			return -EPERM;
		}
	} 
    else 
    {
#ifdef CPMAC_VETH_DEBUG
		printk(KERN_DEBUG "unregister_cpmac_veth_device: Could not find dev\n");
#endif /* CPMAC_VETH_DEBUG */
		return -EINVAL;
	}
}

static int cpmac_veth_device_event(struct notifier_block *unused, 
                                   unsigned long event, 
                                   void *ptr)
{
	int i = 0;
	struct net_device *dev = (struct net_device *)(ptr);
	struct net_device *vdev = NULL;

	switch (event) 
    {
        case NETDEV_CHANGEADDR:
            /* Ignore for now */
            break;

        case NETDEV_GOING_DOWN:
            /* Ignore for now */
            break;

        case NETDEV_DOWN:
        {
			int flgs = 0;
			for (i = 0; i < CPMAC_VETH_DEV_NUM; i++) 
            {
				vdev = cpmac_veth_devices[i];
				if (!vdev ||
				    (CPMAC_VETH_DEV_INFO(vdev)->real_dev != dev) ||
				    (!(vdev->flags & IFF_UP)))
					continue;

				flgs = vdev->flags;
				flgs &= ~IFF_UP;
				dev_change_flags(vdev, flgs);
			}

		    break;
		}

        case NETDEV_UP:
        {
			int flgs = 0;
            for (i = 0; i < CPMAC_VETH_DEV_NUM; i++) 
            {
                vdev = cpmac_veth_devices[i];
                if (!vdev ||
                    (CPMAC_VETH_DEV_INFO(vdev)->real_dev != dev) ||
                    (!(vdev->flags & IFF_UP)))
                    continue;

                flgs = vdev->flags;
                flgs |= IFF_UP;
                dev_change_flags(vdev, flgs);
            }

            break;
        }

	    case NETDEV_UNREGISTER:
        {
            for (i = 0; i < CPMAC_VETH_DEV_NUM; i++) 
            {
                vdev = cpmac_veth_devices[i];
                if (!vdev ||
                    (CPMAC_VETH_DEV_INFO(vdev)->real_dev != dev))
                    continue;

				unregister_cpmac_veth_dev(
					CPMAC_VETH_DEV_INFO(vdev)->real_dev->ifindex,
					CPMAC_VETH_DEV_INFO(vdev)->port_num,
					0, 0);
				vdev = NULL;
            }

            break;
        }
	};

	return NOTIFY_DONE;
}

#endif /* ACTION_TEC_VIRTUAL_ETH */
