#include <linux/types.h>
#include <linux/init.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/if.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_nat_protocol.h>

static int
tcp_in_range(const struct ip_conntrack_tuple *tuple,
             enum ip_nat_manip_type maniptype,
             const union ip_conntrack_manip_proto *min,
             const union ip_conntrack_manip_proto *max)
{
    u_int16_t port;

    if (maniptype == IP_NAT_MANIP_SRC)
        port = tuple->src.u.tcp.port;
    else
        port = tuple->dst.u.tcp.port;

    return ntohs(port) >= ntohs(min->tcp.port)
    && ntohs(port) <= ntohs(max->tcp.port);
}

static int
tcp_unique_tuple(struct ip_conntrack_tuple *tuple,
                 const struct ip_nat_range *range,
                 enum ip_nat_manip_type maniptype,
                 const struct ip_conntrack *conntrack)
{
    static u_int16_t port = 0, *portptr;
    unsigned int range_size, min, i;

    if (maniptype == IP_NAT_MANIP_SRC)
        portptr = &tuple->src.u.tcp.port;
    else
        portptr = &tuple->dst.u.tcp.port;

    /* If no range specified... */
    if (!(range->flags & IP_NAT_RANGE_PROTO_SPECIFIED)) {
        /* If it's dst rewrite, can't change port */
        if (maniptype == IP_NAT_MANIP_DST)
            return 0;

        /* Map privileged onto privileged. */
        if (ntohs(*portptr) < 1024) {
            /* Loose convention: >> 512 is credential passing */
            if (ntohs(*portptr)<512) {
                min = 1;
                range_size = 511 - min + 1;
            } else {
                min = 600;
                range_size = 1023 - min + 1;
            }
        } else {
            min = 1024;
            range_size = 65535 - 1024 + 1;
        }
    } else {
        min = ntohs(range->min.tcp.port);
        range_size = ntohs(range->max.tcp.port) - min + 1;
    }

    for (i = 0; i < range_size; i++, port++) {
        *portptr = htons(min + port % range_size);
        if (!ip_nat_used_tuple(tuple, conntrack)) {
            return 1;
        }
    }
    return 0;
}

static void
esp_manip_pkt(struct iphdr *iph, size_t len,
              const struct ip_conntrack_manip *manip,
              enum ip_nat_manip_type maniptype)
{
    struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
    u_int32_t oldip;
//	u_int16_t *portptr;

    if (maniptype == IP_NAT_MANIP_SRC) {
        /* Get rid of src ip and src pt */
        oldip = iph->saddr;
        ipsec_relative_count_packets(iph->saddr);
        ipsec_add_local_entry(iph->saddr,iph->daddr,ESP_jiffies);
//		printk("*");
//		portptr = &hdr->source;
    } else {
        /* Get rid of dst ip and dst pt */
        oldip = iph->daddr;
//		portptr = &hdr->dest;
    }

    /* this could be a inner header returned in icmp packet; in such
       cases we cannot update the checksum field since it is outside of
       the 8 bytes of transport layer headers we are guaranteed */
    if (((void *)&hdr->check + sizeof(hdr->check) - (void *)iph) <= len) {
        ;//		hdr->check = ip_nat_cheat_check(~oldip, manip->ip,hdr->check);
    }

//we won't change spi	*portptr = manip->u.tcp.port;
}

static unsigned int
tcp_print(char *buffer,
          const struct ip_conntrack_tuple *match,
          const struct ip_conntrack_tuple *mask)
{
    unsigned int len = 0;

    if (mask->src.u.tcp.port)
        len += sprintf(buffer + len, "srcpt=%u ",
                       ntohs(match->src.u.tcp.port));


    if (mask->dst.u.tcp.port)
        len += sprintf(buffer + len, "dstpt=%u ",
                       ntohs(match->dst.u.tcp.port));

    return len;
}

static unsigned int
tcp_print_range(char *buffer, const struct ip_nat_range *range)
{
    if (range->min.tcp.port != 0 || range->max.tcp.port != 0xFFFF) {
        if (range->min.tcp.port == range->max.tcp.port)
            return sprintf(buffer, "port %u ",
                           ntohs(range->min.tcp.port));
        else
            return sprintf(buffer, "ports %u-%u ",
                           ntohs(range->min.tcp.port),
                           ntohs(range->max.tcp.port));
    } else return 0;
}
//Move away later on for pref.
#ifdef ACTION_TEC
struct ipsec ips_ipsec[MAX_IPS_IPSEC];
void init_ips_ipsec()
{
    int i;
    for (i=0;i<MAX_IPS_IPSEC;i++) {
        memset(&ips_ipsec[i],0x0,sizeof(struct ipsec));
		ips_ipsec[i].sequence_number=-1; /*Don't put 0 here, it will increase unneeded triffic.*/
    }
}

void ipsec_relative_count_packets(u_int32_t ip)
{
    int i,j;
    for (i=0;i<MAX_IPS_IPSEC;i++) {
        if (ip == ips_ipsec[i].sip) {
            if (ips_ipsec[i].relative_count_packets++ ==0xfffffffe) {
                for (j=0;j<MAX_IPS_IPSEC;j++)
                    ips_ipsec[j].relative_count_packets=0;
                ips_ipsec[i].relative_count_packets=0x010;
            }

            if (del_timer(&ips_ipsec[i].timeout)) {
                ips_ipsec[i].timeout.expires = jiffies + ESP_jiffies;
                add_timer(&ips_ipsec[i].timeout);
            }
            return ; //already have one.
        }
    }

}

static int sysctl_ipsec_pass_udp=0; /*for avoiding netscreen and w2k conflict.*/
void ipsec_add_local_entry(u_int32_t sip,u_int32_t dip,unsigned long ext_jiffies)
{
    struct ipsec *ipp_ipsec=&ips_ipsec[0];
    int i;
    for (i=0;i<MAX_IPS_IPSEC;i++) {
        if ((ips_ipsec[i].sip == sip) && (ips_ipsec[i].dip == dip)) {
            if ((sysctl_ipsec_pass==0)&&(ext_jiffies==ESP_jiffies)) sysctl_ipsec_pass++;
//            printk("ACTION_TEC DEBUG : already in sysctl_ipsec_pass %x",sysctl_ipsec_pass);
            return; /*already in table, don't add it.*/
        }
    }

	if ((sysctl_ipsec_pass==0)&&(ext_jiffies==ESP_jiffies)) /*No isakmp yet, just leave it.to avoid crash when having netscree and w2k*/
		return;
/*If is not in table yet, find less packets one and replaced it.*/
    for (i=0;i<MAX_IPS_IPSEC;i++) {
        if (ipp_ipsec->relative_count_packets>ips_ipsec[i].relative_count_packets)
            ipp_ipsec=&ips_ipsec[i];
    }
    ipp_ipsec->sip=sip;
    ipp_ipsec->dip=dip;
/*Add timer for auto destroy connection*/

	del_timer (&ipp_ipsec->timeout);
//    init_timer(&ipp_ipsec->timeout);
    ipp_ipsec->timeout.data = (unsigned long)ipp_ipsec;
    if (ext_jiffies==ESP_jiffies){
		sysctl_ipsec_pass++;
		ipp_ipsec->timeout.function = ipsec_death_by_timeout_esp;
	}
	else{
        sysctl_ipsec_pass_udp++;
		ipp_ipsec->timeout.function = ipsec_death_by_timeout_udp;
	}
	ipp_ipsec->timeout.expires = jiffies + ext_jiffies;
    add_timer(&ipp_ipsec->timeout);
    printk("ACTION_TEC DEBUG : add sysctl_ipsec_pass %x",sysctl_ipsec_pass);

}

void ipsec_death_by_timeout_esp(unsigned long ipp_ipsec)
{
    struct ipsec *ipp=(void *)ipp_ipsec;
	
	del_timer (&ipp->timeout);

	ipp->sip=ipp->dip=ipp->relative_count_packets=0;
    if (sysctl_ipsec_pass)
        sysctl_ipsec_pass--;
    printk("ACTION_TEC DEBUG : dec sysctl_ipsec_pass %x",sysctl_ipsec_pass);
}
void ipsec_death_by_timeout_udp(unsigned long ipp_ipsec)
{
    struct ipsec *ipp=(void *)ipp_ipsec;
	del_timer (&ipp->timeout);
    ipp->sip=ipp->dip=ipp->relative_count_packets=0;
    if (sysctl_ipsec_pass_udp)
        sysctl_ipsec_pass_udp--;
    printk("ACTION_TEC DEBUG : dec sysctl_ipsec_pass_udp %x",sysctl_ipsec_pass_udp);
}

#endif




struct ip_nat_protocol ip_nat_protocol_esp
= { { NULL, NULL}, "ESP", IPPROTO_ESP,
    esp_manip_pkt,
    tcp_in_range,
    tcp_unique_tuple,
    tcp_print,
    tcp_print_range
};
