/* vi: set sw=4 ts=4: */
/*
 * Mini syslogd implementation for busybox
 *
 * Copyright (C) 1999,2000 by Lineo, inc. and Erik Andersen
 * Copyright (C) 1999,2000,2001 by Erik Andersen <andersee@debian.org>
 *
 * Copyright (C) 2000 by Karl M. Hegbloom <karlheg@debian.org>
 *
 * "circular buffer" Copyright (C) 2001 by Gennady Feldman <gfeldman@cachier.com>
 *
 * Maintainer: Gennady Feldman <gena01@cachier.com> as of Mar 12, 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <paths.h>
#include <signal.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/param.h>

#include "busybox.h"

/* SYSLOG_NAMES defined to pull some extra junk from syslog.h */
#define SYSLOG_NAMES
#include <sys/syslog.h>
#include <sys/uio.h>


#ifdef ACTION_TEC_PERSISTENT_LOG

#define ACTION_TEC_KPRINTF

#ifdef ACTION_TEC_KPRINTF
        #include <unistd.h>
        #include <sys/file.h>
        #include <time.h>
        #include <stdarg.h>
        #include <fcntl.h>
        #define LED_KPRINTF 8
int kprintf (const char *fmt, ...)
{
        static char tmp_text[51];
        char tmpbuf[128];
        va_list ap;
        int fp;

        /* Open /dev/led file for LED Operation */
        fp = open("/dev/led", O_RDWR);
        if (!fp) {
                return -1;
        }

        memset (tmp_text, 0, sizeof (tmp_text));
        memset (tmpbuf, 0, sizeof(tmpbuf));
        va_start(ap, fmt);
        vsnprintf(tmp_text, sizeof (tmp_text)-1, fmt, ap);
        vsnprintf(tmpbuf, sizeof (tmpbuf)-1, fmt, ap);
        va_end(ap);
        ioctl(fp, LED_KPRINTF, &tmp_text);
        close(fp);

        if (strlen(tmpbuf) >= 50) {
                fp = open("/dev/led", O_RDWR);
                if (!fp) {
                        return -1;
                }
                ioctl(fp, LED_KPRINTF, &tmpbuf[49]);
                close(fp);
        }

        return 0;
}
#endif
#endif

#ifdef ACTION_TEC
    #define MAX_BYTES_OF_FILE_MESSAGE 30720
#endif

/* Path for the file where all log messages are written */
#define __LOG_FILE "/var/log/messages"

/* Path to the unix socket */
static char lfile[BUFSIZ];

static char *logFilePath = __LOG_FILE;


#ifdef ACTION_TEC_PERSISTENT_LOG
#define __PERSISTENT_LOG_FILE "/var/log/persistent_log"
static char *plogFilePath = __PERSISTENT_LOG_FILE;
#endif

/* interval between marks in seconds */
static int MarkInterval = 20 * 60;

/* localhost's name */
static char LocalHostName[32];

#ifdef CONFIG_FEATURE_REMOTE_LOG
    #include <netinet/in.h>
/* udp socket for logging to remote host */
static int remotefd = -1;
/* where do we log? */
    #if defined(ACTION_TEC_PARENTCONTROL) || defined(ACTION_TEC_CMS)
static char *RemoteHost = NULL;
    #else 
static char *RemoteHost;
    #endif //ACTION_TEC_PARENTCONTROL
/* what port to log to? */
static int RemotePort = 514;
/* To remote log or not to remote log, that is the question. */
static int doRemoteLog = FALSE;
static int local_logging = FALSE;
#endif

/* circular buffer variables/structures */
#ifdef CONFIG_FEATURE_IPC_SYSLOG

    #if __GNU_LIBRARY__ < 5
        #error Sorry.  Looks like you are using libc5.
        #error libc5 shm support isnt good enough.
        #error Please disable CONFIG_FEATURE_IPC_SYSLOG
    #endif

    #include <sys/ipc.h>
    #include <sys/sem.h>
    #include <sys/shm.h>

/* our shared key */
static const long KEY_ID = 0x414e4547; /*"GENA"*/

// Semaphore operation structures
static struct shbuf_ds {
    int size;       // size of data written
    int head;       // start of message list
    int tail;       // end of message list
    char data[1];       // data/messages
} *buf = NULL;          // shared memory pointer

static struct sembuf SMwup[1] = {{1, -1, IPC_NOWAIT}}; // set SMwup
static struct sembuf SMwdn[3] = {{0, 0}, {1, 0}, {1, +1}}; // set SMwdn

static int  shmid = -1; // ipc shared memory id
static int  s_semid = -1;   // ipc semaphore id
int data_size = 16000; // data size
int shm_size = 16000 + sizeof(*buf); // our buffer size
static int circular_logging = FALSE;

/*
 * sem_up - up()'s a semaphore.
 */
static inline void sem_up(int semid)
{
    if ( semop(semid, SMwup, 1) == -1 )
        perror_msg_and_die("semop[SMwup]");
}

/*
 * sem_down - down()'s a semaphore
 */
static inline void sem_down(int semid)
{
    if ( semop(semid, SMwdn, 3) == -1 )
        perror_msg_and_die("semop[SMwdn]");
}


void ipcsyslog_cleanup(void){
    printf("Exiting Syslogd!\n");
    if (shmid != -1)
        shmdt(buf);

    if (shmid != -1)
        shmctl(shmid, IPC_RMID, NULL);
    if (s_semid != -1)
        semctl(s_semid, 0, IPC_RMID, 0);
}

void ipcsyslog_init(void){
    if (buf == NULL) {
        if ((shmid = shmget(KEY_ID, shm_size, IPC_CREAT | 1023)) == -1)
            perror_msg_and_die("shmget");


        if ((buf = shmat(shmid, NULL, 0)) == NULL)
            perror_msg_and_die("shmat");


        buf->size=data_size;
        buf->head=buf->tail=0;

        // we'll trust the OS to set initial semval to 0 (let's hope)
        if ((s_semid = semget(KEY_ID, 2, IPC_CREAT | IPC_EXCL | 1023)) == -1) {
            if (errno == EEXIST) {
                if ((s_semid = semget(KEY_ID, 2, 0)) == -1)
                    perror_msg_and_die("semget");
            } else
                perror_msg_and_die("semget");
        }
    } else {
        printf("Buffer already allocated just grab the semaphore?");
    }
}

/* write message to buffer */
void circ_message(const char *msg){
    int l=strlen(msg)+1; /* count the whole message w/ '\0' included */

    sem_down(s_semid);

    /*
     * Circular Buffer Algorithm:
     * --------------------------
     *
     * Start-off w/ empty buffer of specific size SHM_SIZ
     * Start filling it up w/ messages. I use '\0' as separator to break up messages.
     * This is also very handy since we can do printf on message.
     *
     * Once the buffer is full we need to get rid of the first message in buffer and
     * insert the new message. (Note: if the message being added is >1 message then
     * we will need to "remove" >1 old message from the buffer). The way this is done
     * is the following:
     *	When we reach the end of the buffer we set a mark and start from the beginning.
     *	Now what about the beginning and end of the buffer? Well we have the "head"
     *	index/pointer which is the starting point for the messages and we have "tail"
     *	index/pointer which is the ending point for the messages. When we "display" the
     *	messages we start from the beginning and continue until we reach "tail". If we
     *	reach end of buffer, then we just start from the beginning (offset 0). "head" and
     *	"tail" are actually offsets from the beginning of the buffer.
     *
     * Note: This algorithm uses Linux IPC mechanism w/ shared memory and semaphores to provide
     * 	 a threasafe way of handling shared memory operations.
     */
    if ( (buf->tail + l) < buf->size ) {
        /* before we append the message we need to check the HEAD so that we won't
           overwrite any of the message that we still need and adjust HEAD to point
           to the next message! */
        if ( buf->tail < buf->head) {
            if ( (buf->tail + l) >= buf->head ) {
                /* we need to move the HEAD to point to the next message
                 * Theoretically we have enough room to add the whole message to the
                 * buffer, because of the first outer IF statement, so we don't have
                 * to worry about overflows here!
                 */
                int k= buf->tail + l - buf->head; /* we need to know how many bytes
                                 we are overwriting to make
                                 enough room */
                char *c=memchr(buf->data+buf->head + k,'\0',buf->size - (buf->head + k));
                if (c != NULL) {/* do a sanity check just in case! */
                    buf->head = c - buf->data + 1; /* we need to convert pointer to
                                      offset + skip the '\0' since
                                      we need to point to the beginning
                                      of the next message */
                    /* Note: HEAD is only used to "retrieve" messages, it's not used
                        when writing messages into our buffer */
                } else { /* show an error message to know we messed up? */
                    printf("Weird! Can't find the terminator token??? \n");
                    buf->head=0;
                }
            }
        } /* in other cases no overflows have been done yet, so we don't care! */

        /* we should be ok to append the message now */
        strncpy(buf->data + buf->tail,msg,l); /* append our message */
        buf->tail+=l; /* count full message w/ '\0' terminating char */
    } else {
        /* we need to break up the message and "circle" it around */
        char *c;
        int k=buf->tail + l - buf->size; /* count # of bytes we don't fit */

        /* We need to move HEAD! This is always the case since we are going
         * to "circle" the message.
         */
        c=memchr(buf->data + k ,'\0', buf->size - k);

        if (c != NULL) { /* if we don't have '\0'??? weird!!! */
            /* move head pointer*/
            buf->head=c-buf->data+1; 

            /* now write the first part of the message */           
            strncpy(buf->data + buf->tail, msg, l - k - 1);

            /* ALWAYS terminate end of buffer w/ '\0' */
            buf->data[buf->size-1]='\0'; 

            /* now write out the rest of the string to the beginning of the buffer */
            strcpy(buf->data, &msg[l-k-1]);

            /* we need to place the TAIL at the end of the message */
            buf->tail = k + 1;
        } else {
            printf("Weird! Can't find the terminator token from the beginning??? \n");
            buf->head = buf->tail = 0; /* reset buffer, since it's probably corrupted */
        }

    }
    sem_up(s_semid);
}
#endif
/* Note: There is also a function called "message()" in init.c */
/* Print a message to the log file. */
static void message (char *fmt, ...) __attribute__ ((format (printf, 1, 2)));
static void message (char *fmt, ...)
{
    int fd;
    struct flock fl;
    va_list arguments;

    fl.l_whence = SEEK_SET;
    fl.l_start  = 0;
    fl.l_len    = 1;

#ifdef CONFIG_FEATURE_IPC_SYSLOG
    if ((circular_logging) && (buf != NULL)) {
        char b[1024];
        va_start (arguments, fmt);
        vsprintf (b, fmt, arguments);
        va_end (arguments);
        circ_message(b);

    } else
#endif
        if ((fd = device_open (logFilePath,
                               O_WRONLY | O_CREAT | O_NOCTTY | O_APPEND |
                               O_NONBLOCK)) >= 0) {
        fl.l_type = F_WRLCK;
        fcntl (fd, F_SETLKW, &fl);
        va_start (arguments, fmt);
        vdprintf (fd, fmt, arguments);
        va_end (arguments);
        fl.l_type = F_UNLCK;
        fcntl (fd, F_SETLKW, &fl);
        close (fd);

#ifdef ACTION_TEC
        {
            struct stat statbuf;
            if (stat(logFilePath, &statbuf) >= 0) { /*Thisis inpossible*/
                if (statbuf.st_size > MAX_BYTES_OF_FILE_MESSAGE) {
                    unlink (logFilePath);/*Let begin again.*/
                }
            }
        }
#endif
    } else {
        /* Always send console messages to /dev/console so people will see them. */
        if ((fd = device_open (_PATH_CONSOLE,
                               O_WRONLY | O_NOCTTY | O_NONBLOCK)) >= 0) {
            va_start (arguments, fmt);
            vdprintf (fd, fmt, arguments);
            va_end (arguments);
            close (fd);
        } else {
            fprintf (stderr, "Bummer, can't print: ");
            va_start (arguments, fmt);
            vfprintf (stderr, fmt, arguments);
            fflush (stderr);
            va_end (arguments);
        }
    }
}


#ifdef ACTION_TEC_PERSISTENT_LOG
#define LIMIT_OFFSET 100 /*bytes,Used to determine when the file size limit is going to be reached.*/
#define PERSISTENT_LOG_MAX_SIZE (64*1024)
#define FIRST_N_MESG 200
static void p_message (char *fmt, ...) __attribute__ ((format (printf, 1, 2)));
static void p_message (char *fmt, ...)
{
    int fd;
    struct flock fl;
    va_list arguments;

    fl.l_whence = SEEK_SET;
    fl.l_start  = 0;
    fl.l_len    = 1;

#ifdef CONFIG_FEATURE_IPC_SYSLOG
    if ((circular_logging) && (buf != NULL)) {
        char b[1024];
        va_start (arguments, fmt);
        vsprintf (b, fmt, arguments);
        va_end (arguments);
        circ_message(b);

    } else
#endif
        if ((fd = device_open (plogFilePath,
                               O_WRONLY | O_CREAT | O_NOCTTY | O_APPEND |
                               O_NONBLOCK)) >= 0) {

        fl.l_type = F_WRLCK;
        fcntl (fd, F_SETLKW, &fl);
        va_start (arguments, fmt);
        vdprintf (fd, fmt, arguments);
        va_end (arguments);
        fl.l_type = F_UNLCK;
        fcntl (fd, F_SETLKW, &fl);
        close (fd);

#ifdef ACTION_TEC
        {
            FILE *pf;
			char *buf = NULL;
			size_t bytes,size;
			char *pos;
			int line_cnt = 0;
    
            if(!(pf = fopen(__PERSISTENT_LOG_FILE, "r")))
            {
                        kprintf("--not open syslog file\n");
                        return;
            }
            fseek(pf, 0, SEEK_END);
        	size = ftell(pf);
        	rewind(pf);
            if(!size || (size + LIMIT_OFFSET < PERSISTENT_LOG_MAX_SIZE))
            {
               fclose(pf);
            }else           
            {  				
					buf = (char*)malloc(size+1);
					if(!buf)
                    {
						kprintf("--Failed to allocate mem\n");
                        fclose(pf);
						return;					
					}
					memset(buf,0,size+1);
					bytes=fread(buf, 1, size, pf);
                    fclose(pf);
    				
                    if(bytes  <= 0)
					{
                        kprintf("---Failed to read persistent_log file\n");
                        free(buf);
                        return;
                    }else
                    {
						system("echo \"\" > /var/log/persistent_log\n");
						pos = buf;
						while((line_cnt < FIRST_N_MESG) && (*pos != 0))
						{
							if(*pos++ == '\n')
							{
								line_cnt++;
							}
						}
						if(line_cnt == FIRST_N_MESG)  /*we start from No.201 messages and save back to file.*/
						{
							pf = fopen(__PERSISTENT_LOG_FILE, "w");
							if(!pf)
		                    {
        		                kprintf("--not open syslog file\n");
                                free(buf);
                		        return;
                    		}
							fwrite(pos, 1, strlen(pos), pf);
						    fclose(pf);
                            free(buf);
						}else
                        {
                            free(buf);
                            return;
                        }
	               }
            }
        }
#endif
    } else {
        /* Always send console messages to /dev/console so people will see them. */
        if ((fd = device_open (_PATH_CONSOLE,
                               O_WRONLY | O_NOCTTY | O_NONBLOCK)) >= 0) {
            va_start (arguments, fmt);
            vdprintf (fd, fmt, arguments);
            va_end (arguments);
            close (fd);
        } else {
            fprintf (stderr, "Bummer, can't print: ");
            va_start (arguments, fmt);
            vfprintf (stderr, fmt, arguments);
            fflush (stderr);
            va_end (arguments);
        }
    }
}
#endif
#ifdef ACTION_TEC_PERSISTENT_LOG
#define MAX_MSG_BUF_LEN 1024
#endif
static void logMessage (int pri, char *msg)
{
    time_t now;
    char *timestamp;
    static char res[20] = "";
    CODE *c_pri, *c_fac;

    if (pri != 0) {
        for (c_fac = facilitynames;
            c_fac->c_name && !(c_fac->c_val == LOG_FAC(pri) << 3); c_fac++);
        for (c_pri = prioritynames;
            c_pri->c_name && !(c_pri->c_val == LOG_PRI(pri)); c_pri++);
        if (c_fac->c_name == NULL || c_pri->c_name == NULL)
            snprintf(res, sizeof(res), "<%d>", pri);
        else
            snprintf(res, sizeof(res), "%s.%s", c_fac->c_name, c_pri->c_name);
    }

    if (strlen(msg) < 16 || msg[3] != ' ' || msg[6] != ' ' ||
        msg[9] != ':' || msg[12] != ':' || msg[15] != ' ') {
        time(&now);
#if defined(ACTION_TEC) && !defined(ACTION_TEC_VERIZON_TIME_FORMAT)
        timestamp = ctime_offset(&now) + 4;
#else		
        timestamp = ctime(&now) + 4;
#endif		
        timestamp[15] = '\0';
    } else {
        timestamp = msg;
        timestamp[15] = '\0';
        msg += 16;
    }

     
#ifdef CONFIG_FEATURE_REMOTE_LOG
    /* send message to remote logger */
#ifdef ACTION_TEC_REMOTE_SYSLOG
    {
#define FILE_REMOTE_SYSLOG  "/var/tmp/remote_syslog_server"
        struct sockaddr_in remoteaddr;
        FILE *fp;
        char buf[30];
        char *ip;
        char *portstr;
        int  port = 0;

        fp = fopen (FILE_REMOTE_SYSLOG, "r");
        if (fp) {
            memset (buf, 0, sizeof (buf));
            fgets (buf, sizeof (buf) -1, fp);
            ip = strtok (buf, " \n\t");
            if (ip) {
                portstr = strtok (NULL, " \n\t");
                if (portstr) port = atoi (portstr);
                if (!port) port =  RemotePort;

                memset(&remoteaddr, 0, sizeof (struct sockaddr_in));
                remoteaddr.sin_family = AF_INET;
                remoteaddr.sin_port = htons(port);
                inet_pton (AF_INET, ip, &remoteaddr.sin_addr);
                if (remoteaddr.sin_addr.s_addr != 0) {
                    if (remotefd < 0) {
                        remotefd = socket(AF_INET, SOCK_DGRAM, 0);
                    }
                    if (remotefd > 0) {
                        sendto (remotefd, msg, strlen (msg), 0, &remoteaddr, sizeof (struct sockaddr_in ));
                    }
                }
            }
            fclose (fp);
        }
    }
#else //ACTION_TEC_REMOTE_SYSLOG
    if ( -1 != remotefd) {
#if defined(ACTION_TEC_PARENTCONTROL) || defined(ACTION_TEC_CMS)
#define PREFIX_LOG_CMS      "cms_ip: CMS_RePoRt:"
#define PREFIX_REF_CMS      "cms_rf: CMS_RePoRt:"
#define PREFIX_DEFUSER      "cms_defacct: CMS_DeFuSeR:"
#define FILE_REMOTE_SERVER  "/var/etc/pc/remote_server"
        struct sockaddr_in remoteaddr;
        FILE *fp;
        char buf[30];
        char *ip;
        char *pMsg;
        char *portstr;
        int  port = 0;

        if ((strncmp (msg, PREFIX_LOG_CMS, strlen(PREFIX_LOG_CMS)) == 0) || 
            (strncmp (msg, PREFIX_REF_CMS, strlen(PREFIX_REF_CMS)) == 0) ||
            (strncmp (msg, PREFIX_DEFUSER, strlen(PREFIX_DEFUSER)) == 0)) {
            pMsg = msg ; //+ strlen ("syslog: ");
            fp = fopen (FILE_REMOTE_SERVER, "r");
            if (fp) {
                memset (buf, 0, sizeof (buf));
                fgets (buf, sizeof (buf) -1, fp);
                ip = strtok (buf, " \n\t");
                if (ip) {
                    portstr = strtok (NULL, " \n\t");
                    if (portstr) port = atoi (portstr);
                    if (!port) port =  RemotePort;

                    memset(&remoteaddr, 0, sizeof (struct sockaddr_in));
                    remoteaddr.sin_family = AF_INET;
                    remoteaddr.sin_port = htons(port);
                    inet_pton (AF_INET, ip, &remoteaddr.sin_addr);
                    if (remoteaddr.sin_addr.s_addr != 0) {
                        sendto (remotefd, pMsg, strlen (pMsg), 0, &remoteaddr, sizeof (struct sockaddr_in ));
                    }
                }
                fclose (fp);
            }
            return; //don't record cms command to local syslog
        }
#else 
        static const int IOV_COUNT = 2;
        struct iovec iov[IOV_COUNT];
        struct iovec *v = iov;

        memset(&res, 0, sizeof(res));
        snprintf(res, sizeof(res), "<%d>", pri);
        v->iov_base = res ;
        v->iov_len = strlen(res);          
        v++;

        v->iov_base = msg;
        v->iov_len = strlen(msg);          

        writev_retry:
        if ( -1 == writev(remotefd,iov, IOV_COUNT)) {
            if (errno == EINTR) goto writev_retry;
            error_msg_and_die("syslogd: cannot write to remote file handle on"
                              "%s:%d",RemoteHost,RemotePort);
        }
#endif //ACTION_TEC_PARENTCONTROL
    }
#endif //ACTION_TEC_REMOTE_SYSLOG

    if (local_logging)
#endif
        /* now spew out the message to wherever it is supposed to go */
#ifdef ACTION_TEC_VERIZON_TIME_FORMAT
    {
        char buf[30];
        static time_t time_start = 0;
        static int use_GMT = 0;
        time_t now;
        FILE* fp;

		char buf_gmt[20] = {0};
		FILE* fp_gmt;
		int gmt_idx = 0;
#ifdef ACTION_TEC_ADJUST_TIME
        int day_time = 0;
        FILE* fp_ds;
        char buf_ds[20] = {0};
#endif
        memset (buf, 0, sizeof (buf));
        if (time_start == 0){        
            time (&time_start);
#ifdef ACTION_TEC_TR69
{
    FILE *f=NULL;
    f=fopen("/var/tmp/start_t","w");
    if(f){
        fprintf(f,"%ld",time_start);
        fclose(f);
    }
}
#endif

        }

        time(&now);

//      if (strstr (msg, "GoTnTp")) 
//          use_GMT = 1;
        fp = fopen("/var/tmp/offset_t", "r");
        if (fp) {
            long a1=0, a2=0;
            use_GMT = fscanf(fp, "%d %d", &a1, &a2);
#ifndef ACTION_TEC_ADJUST_TIME
            if (use_GMT==2) now+=a1;
#endif
            fclose(fp);
        }
#ifdef ACTION_TEC_QWEST_06
        else
        {
           fp = fopen("/var/tmp/persistent_time", "r");
           if (fp)
           {
               use_GMT=2;
               fclose(fp);
           }
        }
#endif
        if (use_GMT==2) {
			fp_gmt = fopen("/var/tmp/timezone","r");
                                                                                                                             
		    if (fp_gmt != NULL) {
        		if ((fgets(buf_gmt, sizeof(buf_gmt), fp_gmt) >0)) {
	          		gmt_idx = atoi(buf_gmt);
        		}
		        fclose(fp_gmt);
    		}

			memset (buf_gmt, 0, sizeof (buf_gmt));
			if (gmt_idx > 0)
				sprintf(buf_gmt, "GMT+%02d:00", gmt_idx);
			else if (gmt_idx < 0)
				sprintf(buf_gmt, "GMT-%02d:00", abs(gmt_idx));
			else
				sprintf(buf_gmt, "GMT");

#ifdef ACTION_TEC_ADJUST_TIME
            fp_ds = fopen("/var/tmp/daylightsaving","r");
            if(fp_ds!=NULL) 
            {
                if ((fgets(buf_ds, sizeof(buf_ds), fp_ds) >0)) {
                    day_time = atoi(buf_ds);
                }
                else
                    day_time = 0;
                                                                                                                             
                fclose(fp_ds);
            }
            else
                day_time = 0;

            if(day_time != 0)
            {
                char tbuf[50]={0}, strBuf[50]={0};
                FILE *fpx;
                int mYear, mMonth, mWeekday, mMonthday, mHour, mMinute, mSecond, mInt;
                char *bPtr, *ePtr;
                now = now + (gmt_idx*60*60);
                strftime (tbuf, sizeof (tbuf)-1, "%Y-%m-%u-%d-%H-%M-%S", gmtime (&now));
                                                                                                                             
                bPtr = &tbuf[0];
                ePtr = bPtr;
                mInt = 0;
                while ((ePtr != NULL) && (*ePtr))
                {
                    if ((*ePtr) == '-')
                    {
                        switch (mInt)
                        {
                            case 0:
                                *ePtr = 0;
                                strncpy(strBuf, bPtr, strlen(bPtr));
                                mYear = atoi(strBuf);
                                                                                                                             
                                memset(strBuf, 0, sizeof(strBuf));
                                ePtr++;
                                mInt++;
                                if(ePtr)
                                    bPtr = ePtr;
                            break;
                            case 1:
                                *ePtr = 0;
                                strncpy(strBuf, bPtr, strlen(bPtr));
                                mMonth = atoi(strBuf);
                                memset(strBuf, 0, sizeof(strBuf));
                                ePtr++;
                                mInt++;
                                if(ePtr)
                                    bPtr = ePtr;
                            break;
                            case 2:
                                *ePtr = 0;
                                strncpy(strBuf, bPtr, strlen(bPtr));
                                mWeekday = atoi(strBuf);
                                memset(strBuf, 0, sizeof(strBuf));
                                ePtr++;
                                mInt++;
                                if(ePtr)
                                    bPtr = ePtr;
                                break;
                            case 3:
                                *ePtr = 0;
                                strncpy(strBuf, bPtr, strlen(bPtr));
                                mMonthday = atoi(strBuf);
                                memset(strBuf, 0, sizeof(strBuf));
                                ePtr++;
                                mInt++;
                                if(ePtr)
                                    bPtr = ePtr;
                                break;
                            case 4:
                                *ePtr = 0;
                                strncpy(strBuf, bPtr, strlen(bPtr));
                                mHour = atoi(strBuf);
                                memset(strBuf, 0, sizeof(strBuf));
                                ePtr++;
                                mInt++;
                                if(ePtr)
                                    bPtr = ePtr;
                                break;
                            case 5:
                                *ePtr = 0;
                                strncpy(strBuf, bPtr, strlen(bPtr));
                                mMinute = atoi(strBuf);
                                memset(strBuf, 0, sizeof(strBuf));
                                ePtr++;
                                mInt++;
                                if(ePtr)
                                    bPtr = ePtr;
                                break;
                        }
                    } else
                        ePtr++;
                }
                                                                                                                             
                if(( mInt == 6) && (ePtr == NULL) && (bPtr != NULL))
                {
                    strncpy(strBuf, bPtr, strlen(bPtr));
                    mSecond = atoi(strBuf);
                }
                if((mMonth == 1) || (mMonth == 2) || (mMonth == 3) || (mMonth == 11) || (mMonth == 12))
                {
                    day_time = 0;
                } else if((mMonth == 4) && ((mMonthday >= 1) && (7 >= mMonthday)))
                {
                    if ((mWeekday >= 1) && (6 >= mWeekday))
                    {
                        for (mInt = 1; mInt <= 6; mInt++)
                        {
                            if((mWeekday == mInt) && (mMonthday <= mInt))
                                day_time = 0;
                        }
                    } else if((mWeekday == 7) && (mMonthday <= 7))
                    {
                        if(mHour < 2)
                            day_time = 0;
                    }
                } else if((mMonth == 10) && ((mMonthday >= 25) && (31 >= mMonthday)))
                {
                    if ((mWeekday >= 1) && (6 >= mWeekday))
                    {
                        for (mInt = 1; mInt <= 6; mInt++)
                        {
                            if((mWeekday == mInt) && (mMonthday >= (25 + mInt)))
                                day_time = 0;
                        }
                    } else if((mWeekday == 7) && (mMonthday >= 25))
                    {
                        if(mHour >= 2)
                            day_time = 0;
                    }
                }
            } else
                now = now + (gmt_idx*60*60);
                                                                                                                             
            now = now + (day_time*60*60);
#endif
            strftime (buf, sizeof (buf)-1, "%H:%M:%S %a %b %d %Y", gmtime (&now));
#ifdef ACTION_TEC_PERSISTENT_LOG
            if (pri > 100)
            {
				static char last_msg[MAX_MSG_BUF_LEN] = {0};
    			int str_len;
				static long int last_sec = 0; /*timestample of latest log in seconds*/
                long int cur_sec;
                time_t tsec = time(NULL);
                struct tm t; 
                struct tm *tm;
                char pbuf[64];
                char *pos = strchr(msg, ':');
		       if(!pos)
     		   {
           			pos = msg;
     		   }else
     		   {
        			pos++;
     		   }
               /*Here we are gonna drop the duplicate message*/
     		   str_len = strlen(pos);
     		   if(str_len < MAX_MSG_BUF_LEN)
     		   {
     	            strptime(buf,"%H:%M:%S %a %b %d %Y",&t);
	                tsec = mktime(&t);
                    cur_sec =(long int)tsec;
                    //kprintf("DEbug info: %ld",cur_sec);
        			if((!strncmp(last_msg, pos, str_len)) && (cur_sec > last_sec) && (cur_sec-last_sec < 60))
        			{
                        //kprintf("duplicate message within 60 seconds and drop it now\n");
            			return;
        			}else
        			{
                        last_sec = cur_sec;
            			strncpy(last_msg, pos, str_len);
            			last_msg[str_len] = 0;
        			}
     			}
 
                memset(pbuf, 0, sizeof(pbuf));
                tm = gmtime (&now);
                if (tm->tm_hour > 12)
                    sprintf(pbuf, "%02d/%02d/%04d\t%02d:%02d:%02dPM", tm->tm_mon+1, tm->tm_mday, tm->tm_year+1900, tm->tm_hour%12, tm->tm_min, tm->tm_sec);
                else if (tm->tm_hour == 12)
                    sprintf(pbuf, "%02d/%02d/%04d\t%02d:%02d:%02dPM", tm->tm_mon+1, tm->tm_mday, tm->tm_year+1900, tm->tm_hour, tm->tm_min, tm->tm_sec);
                else
                    sprintf(pbuf, "%02d/%02d/%04d\t%02d:%02d:%02dAM", tm->tm_mon+1, tm->tm_mday, tm->tm_year+1900, tm->tm_hour, tm->tm_min, tm->tm_sec);
                //strftime (pbuf, sizeof (pbuf)-1, "%m/%d/%Y\t%H:%M:%S", tm);

                switch(pri)
                {
                   case 101:
                   {  
		                 p_message("%s\tReboot\t%s\n", pbuf, pos);
				         break;
                   }
				   case 102:
 				   {
                         p_message("%s\tLAN DHCP\t%s\n", pbuf, pos);
					     break;
				   }
                   case 109:  /*WAN PPP Authentication Failed Log */
				   case 103:
				   {
                         p_message("%s\tWAN Connection\t%s\n", pbuf, pos);
					  	 break;
				   }
				   case 104:
				   {
                         p_message("%s\tCWMP\t%s\n", pbuf, pos);
						 break;
				   }
				   case 105:
	               {
                         p_message("%s\tCWMP Firmware Upgrade\t%s\n", pbuf, pos);
					     break;
	               }	
				   case 106:
				   {
                         p_message("%s\tLocal Firmware Upgrade\t%s\n", pbuf, pos);
				         break;
				   }
				   case 107:
				   {
                         p_message("%s\tFailed Login\t%s\n", pbuf, pos);
					  	 break;
			       }
				   case 108:
                   {
                         p_message("%s\tDSL Link\t%s\n", pbuf, pos);
                      	 break;
                   }
				   default:
                   {
                         p_message("%s\tUndefined\t%s\n", pbuf, pos);
					  	 break;
                   }
                }
            }
            else
#endif
			message("(%s)%s %s\n", buf_gmt, buf, msg);


            //strftime (buf, sizeof (buf)-1, "%H:%M:%S GMT %a %b %d %Y", gmtime (&now));
        } else {
            long sec_passed;

            sec_passed = now - time_start;
            sprintf(buf, "%5d:%02d:%02d Elapsed Time", 
                    sec_passed/3600, 
                    (sec_passed%3600)/60,
                    sec_passed%60);
#ifdef ACTION_TEC_PERSISTENT_LOG
            if (pri > 100)
                p_message("%s %s\n", buf, msg);
            else
#endif
			message("%s %s\n", buf, msg);
        }
    }
#else
        message("%s %s %s %s\n", timestamp, LocalHostName, res, msg);
#endif ACTION_TEC


}

static void quit_signal(int sig)
{
    logMessage(LOG_SYSLOG | LOG_INFO, "System log daemon exiting.");
    unlink(lfile);
#ifdef CONFIG_FEATURE_IPC_SYSLOG
    ipcsyslog_cleanup();
#endif

    exit(TRUE);
}

static void domark(int sig)
{
    if (MarkInterval > 0) {
#ifndef ACTION_TEC
        logMessage(LOG_SYSLOG | LOG_INFO, "-- MARK --");
#endif
        alarm(MarkInterval);
    }
}

/* This must be a #define, since when DODEBUG and BUFFERS_GO_IN_BSS are
 * enabled, we otherwise get a "storage size isn't constant error. */
#ifndef ACTION_TEC_CMS
    #define BUFSIZE 1023
#else
    #define BUFSIZE 20480
#endif
static int serveConnection (int conn)
{
    RESERVE_CONFIG_BUFFER(tmpbuf, BUFSIZE + 1);
    int    n_read;
    char *p = tmpbuf;

    n_read = read (conn, tmpbuf, BUFSIZE );

    while (p < tmpbuf + n_read) {

        int           pri = (LOG_USER | LOG_NOTICE);
        char          line[ BUFSIZE + 1 ];
        unsigned char c;

        char *q = line;

        tmpbuf[ n_read - 1 ] = '\0';

        while (p && (c = *p) && q < &line[ sizeof (line) - 1 ]) {
#ifdef ACTION_TEC_CMS
            if (c == '<' && !isdigit(*(p+1)) ) {
                *q++ = c;
            } else
#endif
                if (c == '<') {
                /* Parse the magic priority number. */
                pri = 0;
                while (isdigit (*(++p))) {
                    pri = 10 * pri + (*p - '0');
                }
                if (pri & ~(LOG_FACMASK | LOG_PRIMASK)) {
                    pri = (LOG_USER | LOG_NOTICE);
                }
            } else if (c == '\n') {
#ifndef ACTION_TEC_CMS
                *q++ = ' ';   
#endif 
            } else if (iscntrl (c) && (c < 0177)) {
#ifndef ACTION_TEC_CMS
                *q++ = '^';
                *q++ = c ^ 0100;
#endif
#ifdef ACTION_TEC_CMS
            } else if (c == '\t') {
                //filter out \t!
#endif
            } else {
                *q++ = c;
            }
            p++;
        }
        *q = '\0';
        p++;
        /* Now log it */
        logMessage (pri, line);
    }
    RELEASE_CONFIG_BUFFER (tmpbuf);
    return n_read;
}


#ifdef CONFIG_FEATURE_REMOTE_LOG
static void init_RemoteLog (void){

    struct sockaddr_in remoteaddr;
    struct hostent *hostinfo;
    int len = sizeof(remoteaddr);
    memset(&remoteaddr, 0, len);

    remotefd = socket(AF_INET, SOCK_DGRAM, 0);

    if (remotefd < 0) {
        error_msg_and_die("syslogd: cannot create socket");
    }
#if !defined(ACTION_TEC_PARENTCONTROL) && !defined(ACTION_TEC_CMS)
    hostinfo = xgethostbyname(RemoteHost);

    remoteaddr.sin_family = AF_INET;
    remoteaddr.sin_addr = *(struct in_addr *) *hostinfo->h_addr_list;
    remoteaddr.sin_port = htons(RemotePort);

    /* 
       Since we are using UDP sockets, connect just sets the default host and port 
       for future operations
    */
    if ( 0 != (connect(remotefd, (struct sockaddr *) &remoteaddr, len))) {
        error_msg_and_die("syslogd: cannot connect to remote host %s:%d", RemoteHost, RemotePort);
    }
#endif //ACTION_TEC_PARENTCONTROL
}
#endif

#ifdef ACTION_TEC
    #if defined(CONFIG_SYSLOG_SAVE) || defined(ACTION_TEC_SYSLOG_SAVE_INTERVAL)

        #include <sched.h>
        #include <stdio.h>
        #include <stdarg.h>
        #define DEBGU 0 
int kprintf (const char *fmt, ...)
{
#if DEBUG	
#define MAX_KMSGLEN 40 //defined in led driver
    static char tmp_text[MAX_KMSGLEN+3];
    va_list ap;
    FILE *fproc;

    memset (tmp_text, 0, sizeof (tmp_text));
    va_start(ap, fmt);
    vsnprintf(tmp_text+3, sizeof (tmp_text)-3-1, fmt, ap);
    va_end(ap);
    tmp_text[0] = '9';
    tmp_text[1] = '9';
    tmp_text[2] = ',';
    fproc = fopen("/proc/led_mod/led", "w+");

    if (!fproc) {
        return -1;
    }
    fwrite(tmp_text, 1, strlen (tmp_text), fproc);
    fclose(fproc);
#endif	
    return 0;
}

        #define LOG_FILES_NUM 		2
        #define ENTRYSIZE           300

int generate_each_log(int src, const char* filename, int iNo)
{
    FILE *fdst=0, *fsrc;
    char label[256]={0};
    char buf[ENTRYSIZE]={0};
    int rd, wrote=0;
    unsigned int usize = 64/LOG_FILES_NUM*1024;

    lseek(src, iNo*usize, SEEK_SET);
    snprintf(label, 256, "**MARK%d**: %s **\n", iNo, filename);
    read(src, buf, 256);

    if ( strncmp(buf, label, 256) ) {
        kprintf("\n\n\nLog flag not match.\n\n\n");
        return -1;
    }

    fsrc = fdopen(src, "r");
    if (fsrc==0) return -1;

    fdst = fopen(filename, "a+");   
    if (fdst) {
        int lines=0;
        while ( fgets(buf, ENTRYSIZE, fsrc) ) {
            rd = strlen(buf);
            wrote += rd;
            if ( (wrote>=usize) || (wrote>=MAX_BYTES_OF_FILE_MESSAGE) || (rd==0) )
                break;
            lines++;
            fwrite(buf, 1, rd, fdst);
        }   
        if ( lines )
            fprintf(fdst, "%s\n", "The above lines are reloaded from flash.");
        kprintf("\n%d bytes restored, interval is %d minutes.\n", wrote, ACTION_TEC_SYSLOG_SAVE_INTERVAL);
        fclose(fdst);                       
    }

    return 0;
}   

int load_log()
{
    int src=0;

    src = open("/dev/mtdblock/4", O_RDONLY);
    if ( src==-1 ) {
        kprintf("Cannot open mtd4, return...\n");
        return -1;
    }

    generate_each_log(src, logFilePath, 0);
    generate_each_log(src, "/var/log/firewalling", 1);

    close(src);
    return 0;
}   

int save_each_log(int dst, const char* filename, int sizelimit, int iNo)
{
    char buf[ENTRYSIZE]={0};
    FILE* src;
    int rd = 0, wr = 0;
    unsigned long size = 0, offset, wrote=0, iread=0;
    struct stat sz;
    int bSkiped=0;
    unsigned int usize = 64/LOG_FILES_NUM*1024;

    src = fopen(filename, "r");
    if ( src==NULL )
        return -1;

    stat(filename, &sz);
    size = sz.st_size;

    lseek(dst, iNo*usize, SEEK_SET);
    snprintf(buf, 256, "**MARK%d**: %s **\n", iNo, filename);
    write(dst, buf, 256);

    while ( fgets(buf, ENTRYSIZE, src) ) {
        rd = strlen(buf);
        iread += rd;
        if ( sizelimit<size ) {
            if ( (iread<size-sizelimit) || (bSkiped== 0) ) {
                if (bSkiped==0) bSkiped=1;
                memset(buf, 0, ENTRYSIZE);
                continue;
            }
        }

        /* if ((wrote + rd) >= sizelimit)
            rd = sizelimit- wrote;
            if (rd < 0)
            rd = 0;	*/

        offset = 0;
        while (((wr = write(dst, &buf[offset], (rd - offset))) > 0) && (rd - offset))
            offset += wr;
        wrote += offset;

        memset(buf, 0, ENTRYSIZE);
    }           

    {
        char end[10] = {0};
        end[9]=-1;
        write(dst, &end, 10);
    }           
    fclose(src);

    if (wrote)
        kprintf("[%d] %d bytes written.\n", getpid(), wrote);

    return 0;
}   

void save_log(void* ptr)
{
#define LOG_SAVE_LIMIT	  10240

    int dst=0, ilen=0;
    unsigned int i_syslog_saved_size;
    unsigned int i_fwlog_saved_size;

    FILE* fp = fopen("/var/tmp/log_length", "r");
    if (fp) {
        ilen = fscanf(fp, "%d %d", &i_syslog_saved_size, &i_fwlog_saved_size);
        i_syslog_saved_size = i_syslog_saved_size*1024;
        i_fwlog_saved_size  = i_fwlog_saved_size*1024;
        fclose(fp);
    }

    if ( ilen==-1 || ilen==0 ) {
        i_syslog_saved_size = LOG_SAVE_LIMIT;
        i_fwlog_saved_size  = LOG_SAVE_LIMIT;
    }

    if ( (i_syslog_saved_size+i_fwlog_saved_size)==0 )
        return;

    while (1) {
        sleep(ACTION_TEC_SYSLOG_SAVE_INTERVAL*60);

        dst = open("/dev/mtdblock/4", O_WRONLY);
        if ( dst==-1 ) {
            kprintf("Cannot open mtd4, thread exits...\n");
            break;
        }

        if (i_syslog_saved_size) save_each_log(dst, logFilePath, i_syslog_saved_size, 0);
        if (i_fwlog_saved_size) save_each_log(dst, "/var/log/firewalling", i_fwlog_saved_size, 1);

        close(dst);
    }   
}                                                       

    #endif
#endif

static void doSyslogd (void) __attribute__ ((noreturn));
static void doSyslogd (void)
{
    struct sockaddr_un sunx;
    socklen_t addrLength;


    int sock_fd;
    fd_set fds;

    /* Set up signal handlers. */
    signal (SIGINT,  quit_signal);
    signal (SIGTERM, quit_signal);
    signal (SIGQUIT, quit_signal);
    signal (SIGHUP,  SIG_IGN);
    signal (SIGCHLD,  SIG_IGN);
#ifdef SIGCLD
    signal (SIGCLD,  SIG_IGN);
#endif
    signal (SIGALRM, domark);
    alarm (MarkInterval);

    /* Create the syslog file so realpath() can work. */
    if (realpath (_PATH_LOG, lfile) != NULL)
        unlink (lfile);

    memset (&sunx, 0, sizeof (sunx));
    sunx.sun_family = AF_UNIX;
    strncpy (sunx.sun_path, lfile, sizeof (sunx.sun_path));
    if ((sock_fd = socket (AF_UNIX, SOCK_STREAM, 0)) < 0)
        perror_msg_and_die ("Couldn't get file descriptor for socket " _PATH_LOG);

    addrLength = sizeof (sunx.sun_family) + strlen (sunx.sun_path);
    if ((bind (sock_fd, (struct sockaddr *) &sunx, addrLength)) || (listen (sock_fd, 5)))
        perror_msg_and_die ("Could not connect to socket " _PATH_LOG);

    if (chmod (lfile, 0666) < 0)
        perror_msg_and_die ("Could not set permission on " _PATH_LOG);

    FD_ZERO (&fds);
    FD_SET (sock_fd, &fds);

#ifdef CONFIG_FEATURE_IPC_SYSLOG
    if (circular_logging ) {
        ipcsyslog_init();
    }
#endif

#ifdef CONFIG_FEATURE_REMOTE_LOG
    if (doRemoteLog) {
        init_RemoteLog();
    }
#endif
#ifdef ACTION_TEC_VERIZON_TIME_FORMAT
    logMessage (LOG_SYSLOG | LOG_INFO, "syslogd started: BusyBox v0.61.pre" );
#else
    logMessage (LOG_SYSLOG | LOG_INFO, "syslogd started: " BB_BANNER);
#endif 

    for (;;) {

        fd_set readfds;
        int    n_ready;
        int    fd;

        memcpy (&readfds, &fds, sizeof (fds));

        if ((n_ready = select (FD_SETSIZE, &readfds, NULL, NULL, NULL)) < 0) {
            if (errno == EINTR) continue; /* alarm may have happened. */
            perror_msg_and_die ("select error");
        }

        for (fd = 0; (n_ready > 0) && (fd < FD_SETSIZE); fd++) {
            if (FD_ISSET (fd, &readfds)) {

                --n_ready;

                if (fd == sock_fd) {
                    int   conn;

                    //printf("New Connection request.\n");
                    if ((conn = accept (sock_fd, (struct sockaddr *) &sunx, &addrLength)) < 0) {
                        perror_msg_and_die ("accept error");
                    }

                    FD_SET(conn, &fds);
                    //printf("conn: %i, set_size: %i\n",conn,FD_SETSIZE);
                } else {
                    //printf("Serving connection: %i\n",fd);
                    if ( serveConnection(fd) <= 0 ) {
                        close (fd);
                        FD_CLR(fd, &fds);
                    }
                } /* fd == sock_fd */
            }/* FD_ISSET() */
        }/* for */
    } /* for main loop */
}

extern int syslogd_main(int argc, char **argv)
{
    int opt;
    int doFork = TRUE;
    char *p;

    /* do normal option parsing */
    while ((opt = getopt(argc, argv, "m:nO:R:LC")) > 0) {
        switch (opt) {
        case 'm':
            MarkInterval = atoi(optarg) * 60;
            break;
        case 'n':
            doFork = FALSE;
            break;
        case 'O':
            logFilePath = xstrdup(optarg);
            break;
#ifdef CONFIG_FEATURE_REMOTE_LOG
        case 'R':
            RemoteHost = xstrdup(optarg);
            if ( (p = strchr(RemoteHost, ':'))) {
                RemotePort = atoi(p+1);
                *p = '\0';
            }
            doRemoteLog = TRUE;
            break;
        case 'L':
            local_logging = TRUE;
            break;
#endif
#ifdef CONFIG_FEATURE_IPC_SYSLOG
        case 'C':
            circular_logging = TRUE;
            break;
#endif
        default:
            show_usage();
        }
    }

#if defined(ACTION_TEC_PARENTCONTROL) || defined(ACTION_TEC_CMS)
    if (!RemoteHost) RemoteHost = xstrdup("remotehost");  //parent control won't use this var
    doRemoteLog = TRUE;
    local_logging = TRUE;
#endif //ACTION_TEC_PARENTCONTROL
#ifdef CONFIG_FEATURE_REMOTE_LOG
    /* If they have not specified remote logging, then log locally */
    if (! doRemoteLog)
        local_logging = TRUE;
#endif
    /* Store away localhost's name before the fork */
    gethostname(LocalHostName, sizeof(LocalHostName));
    if ((p = strchr(LocalHostName, '.'))) {
        *p++ = '\0';
    }

    umask(0);

    if (doFork) {
#if !defined(__UCLIBC__) || defined(__UCLIBC_HAS_MMU__)
        if (daemon(0, 1) < 0)
            perror_msg_and_die("daemon");
#else
        error_msg_and_die("daemon not supported");
#endif
    }
#ifdef ACTION_TEC
#if defined(CONFIG_SYSLOG_SAVE) || defined(ACTION_TEC_SYSLOG_SAVE_INTERVAL)
    if ( load_log()!=-1 ) {
        if ( fork()==0 ) {
            nice(20);
            save_log(NULL);
            exit(0);
        }
    }
#endif
#endif	
    doSyslogd();

    return EXIT_SUCCESS;
}

/*
Local Variables
c-file-style: "linux"
c-basic-offset: 4
tab-width: 4
End:
*/
