/*
 * tc_core.c		TC core library.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <fcntl.h>
#include <math.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#include "tc_core.h"

#define ACTION_TEC_ADSL_OVERHEAD 1

static __u32 t2us=1;
static __u32 us2t=1;
static double tick_in_usec = 1;

long tc_core_usec2tick(long usec)
{
	return usec*tick_in_usec;
}

long tc_core_tick2usec(long tick)
{
	return tick/tick_in_usec;
}

unsigned tc_calc_xmittime(unsigned rate, unsigned size)
{
	return tc_core_usec2tick(1000000*((double)size/rate));
}

#ifdef ACTION_TEC_ADSL_OVERHEAD
#ifndef ATM_CELL_SIZE
#define ATM_CELL_SIZE               53  /* ATM cell size incl. header */
#endif
#ifndef ATM_CELL_PAYLOAD
#define ATM_CELL_PAYLOAD            48  /* ATM payload size */
#endif

#define ATM_ALIGN y

/*
 The align_to_cells is used for determining the (ATM) SAR alignment
 overhead at the ATM layer. (SAR = Segmentation And Reassembly)
 This is for example needed when scheduling packet on an ADSL
 connection.  The ATM-AAL overhead should preferably be added in the
 kernel when doing table lookups (due to precision/alignment of the
 table), if not the ATM-AAL overhead should be added to the size
 before calling the function.  --Hawk, d.7/11-2004. <hawk@diku.dk>
 */
unsigned tc_align_to_cells(unsigned size, int cell_size, int cell_payload) 
{
  int linksize, cells;
  cells = size / cell_payload;
  if ((size % cell_payload) > 0) {
    cells++;
  }
  linksize = cells * cell_size;
  return linksize;
}

#endif
/*
   rtab[pkt_len>>cell_log] = pkt_xmit_time
 */

int tc_calc_rtable(unsigned bps, __u32 *rtab, int cell_log, unsigned mtu,
		   unsigned mpu)
{
	int i;
#ifdef ACTION_TEC_ADSL_OVERHEAD
	unsigned overhead = (mpu >> 8) & 0xFF;
	mpu = mpu & 0xFF;
#endif

	if (mtu == 0)
		mtu = 2047;

	if (cell_log < 0) {
		cell_log = 0;
		while ((mtu>>cell_log) > 255)
			cell_log++;
	}
	for (i=0; i<256; i++) {
#ifdef ACTION_TEC_ADSL_OVERHEAD
		unsigned sz = ((i+1)<<cell_log);
		if (overhead) {
			// Is now done in the kernel (eg. sch_htb.c, func L2T )
			// sz += overhead; 
		}
		if (sz < mpu)
			sz = mpu;
#ifdef ATM_ALIGN
		sz = tc_align_to_cells(sz, ATM_CELL_SIZE, ATM_CELL_PAYLOAD);
#endif
#else
		unsigned sz = (i<<cell_log);
		if (sz < mpu)
			sz = mpu;
#endif
		rtab[i] = tc_core_usec2tick(1000000*((double)sz/bps));
	}
	return cell_log;
}

int tc_core_init()
{
	FILE *fp = fopen("/proc/net/psched", "r");

	if (fp == NULL)
		return -1;

	if (fscanf(fp, "%08x%08x", &t2us, &us2t) != 2) {
		fclose(fp);
		return -1;
	}
	fclose(fp);
	tick_in_usec = (double)t2us/us2t;
	return 0;
}
