/*
  net-acct --- a network accounting system for Embedded Linux
  Copyright (C) 2004 Steven Jiang

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


  interprocess communication and message lib

*/

#include <unistd.h>
#include <stdio.h>
#include <syslog.h>
#include <malloc.h>
#include <time.h>
#include <errno.h>
#include "comm.h"
#include "im_util.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif /* DMALLOC */

int serv_socket = -1;

static int send_flags = 0;

int comm_start(char *name, int send_fl)
{
	struct sockaddr_un sun;
	int srvsock;

	openlog(name, 0, -1);

	srvsock = socket(AF_UNIX, SOCK_DGRAM, 0);
	if(srvsock == -1) {
		syslog(LOG_ERR, "Low level socket error");
		return COMM_FAIL_SOCK;
	}
	sun.sun_family = AF_UNIX;

#if defined(ACTION_TEC_CMS) || defined(ACTION_TEC_VERIZON_WRA)
	snprintf(sun.sun_path, UNIX_PATH_MAX, "%s/cm_%s.ctl", "/var/tmp", name);
#else //PARENTAL_CONTROL
	snprintf(sun.sun_path, UNIX_PATH_MAX, "%s/parentcontrol_%s.ctl", SOCKET_PATH, name);
#endif
//    printf("before binding:%s/acct_%s.ctl\n", SOCKET_PATH, name);
	if(bind(srvsock, (struct sockaddr*)&sun, sizeof(struct sockaddr_un))<0) {
		/* Ok, maybe the socket already exists  try connecting
		 * to see if another instance is present
		 */
		if(connect(srvsock, (struct sockaddr*)&sun,
				sizeof(struct sockaddr_un))==0){
			syslog(LOG_ERR, "Second instance already running");
            printf ("Second instance already running\n");

			return COMM_FAIL_SECOND_INSTANCE;
		}
		/* That wasn't it, lets try removing the socket from the filesystem */
		if(unlink(sun.sun_path) <0){
			syslog(LOG_ERR, "Error removing old socket");
//            printf("Error removing old socket %s\n", sun.sun_path);
			return COMM_FAIL_UNLINK;
		}
		/* Ok, if we are here, then we unlinked the old socket. Lets bind again */
		if(bind(srvsock, (struct sockaddr*)&sun, sizeof(struct sockaddr_un))<0){
			syslog(LOG_ERR, "Error binding socket");
//            printf ("Error binding socket");
			return COMM_FAIL_BIND;
		}
	}
	serv_socket = srvsock;
	send_flags = send_fl;
	return srvsock;
}

int comm_unlink(char *name)
{
	char path[UNIX_PATH_MAX];

	snprintf(path, UNIX_PATH_MAX, "%s/parentcontrol_%s.ctl", SOCKET_PATH, name);
	return unlink(path);
}

int comm_sendto(int fd, const void *msg, size_t len, char *name)
{
	struct sockaddr_un	to;

	to.sun_family=AF_UNIX;
	snprintf(to.sun_path, UNIX_PATH_MAX, "%s/parentcontrol_%s.ctl", SOCKET_PATH, name);

	return sendto(fd, msg, len, send_flags, (struct sockaddr *)&to, sizeof(struct sockaddr_un));
}

#if !defined(ACTION_TEC_CMS) && !defined(ACTION_TEC_VERIZON_WRA)

void msg_free (struct msg_base *msg){
    if (msg) {
        if (msg->data ) 
            free (msg->data);
        free (msg);
    }
}

int	msg_makebuf (char *buf, int len, struct msg_base *msg)
{
	if(!buf)
		return 0;
	if(!msg)
		return 0;
	if (len < sizeof (struct msg_base) + msg->datalen)
	{
		/* Not enough room in buffer*/
		return -1;
	}

    memcpy (buf, (char *)msg, sizeof (struct msg_base));
    memcpy (buf + sizeof (struct msg_base), msg->data, msg->datalen);

	return sizeof (struct msg_base) + msg->datalen;

}

int msg_sendOut(struct msg_base *msg, char *name)
{
    int size, res;
    static char buf[MAX_MESSAGE_LEN];

    if(serv_socket < 0)
            return -1;

    size = msg_makebuf(buf, sizeof(buf), msg);
    res = comm_sendto(serv_socket, buf, size, name);
    if(res < 0)
    {
        if(errno == EAGAIN)
                return -1;
        else
        {
               syslog (LOG_WARNING, "Failed to deliver message '%d' from '%s' to '%s'",
                                                msg->msg_type, msg->from, msg->to);
                return -1;
        }
    }
    else
        return res;
}

int current_seq = 0; // current sequence number
extern struct msg_base *message_queue;

int msg_getNextSeq()
{
	current_seq++;
	if(current_seq < 0)
		current_seq = 1;
	return current_seq;
}

struct msg_base *new_msg (int msg_type, char *data, int datalen, char *to, char *from)
{
    struct msg_base *msg;

    msg = (struct msg_base *)malloc (sizeof (struct msg_base));
    if (!msg) {
        syslog (LOG_ERR, "Can't get mem for msg_base");
        return NULL;
    }

    memset ((char *)msg, 0, sizeof (struct msg_base));
    strncpy (msg->to, to, sizeof (msg->to) -1);
    strncpy (msg->from, from, sizeof (msg->from) -1);

    msg->data = (char *)malloc (datalen);
    if (!msg->data) {
        syslog (LOG_ERR, "Can't get mem for msg_base data");
        free (msg);
        return NULL;
    }
    memcpy (msg->data, data, datalen);
    msg->datalen = datalen;
    msg->seq = msg_getNextSeq ();
    msg->msg_type = msg_type;
    msg->expire = time(0) + SECONDS_EXPIRE_VERDICT;
    return msg;
}

void msg_flush(void)
{
	struct msg_base *msg;

	while(message_queue)
	{
		msg = message_queue;
		NODE_DEL(message_queue, msg);
		msg_free(msg);
	}
}

int msg_send(struct msg_base *msg)
{
	int ret;

//	printf("Enqueueing message\n");
	if(!msg)
		return -1;

    ret = msg_sendOut(msg, msg->to);
	return 0;
}

int	msg_iterate_queue()
{
	struct msg_base *msg;

	while(message_queue)
	{
		msg = message_queue;
        if(time(0) < msg->expire) // not expired?
			return -1;                // keep in queue then
		
		NODE_DEL(message_queue, msg);
		msg_free(msg);
	}
	return 0;
}

struct msg_base* find_msg (int seq)
{
	struct msg_base *msg = message_queue;

	while(msg)
	{
        if(msg->seq == seq) {
            NODE_DEL(message_queue, msg);
            return msg;
        }
        msg = msg ->next;
	}
	return NULL;
}

#endif //ACTION_TEC_CMS
