/* Shared library add-on to iptables to add sLog support.
 * 
 * (C) 2003 by Mingfei Peng <mpeng@actiontec.com>
 *
 *
 * This software is released under the terms of GNU GPL
 * 
 * libipt_sLog.c,v 1.0 2003/07/31 11:55:02 laforge Exp
 */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <getopt.h>
#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_sLog.h>


#define IPT_SLOG_OPT_MAX_NUM 0x01
#define IPT_SLOG_OPT_TIMEOUT 0x02


/* Function which prints out usage message. */
static void help(void)
{
	printf("sLog v%s options:\n"
	       " --sLog-max_num num		Maximum number of entries\n"
	       " --sLog-timeout timeout		timeout value(in secs)\n\n",
	       IPTABLES_VERSION);
}

static struct option opts[] = {
	{ .name = "sLog-max_num", .has_arg = 1, .flag = 0, .val = '1'},
	{ .name = "sLog-timeout", .has_arg = 1, .flag = 0, .val = '2'},
	{ .name = 0 }
};

/* Initialize the target. */
static void init(struct ipt_entry_target *t, unsigned int *nfcache)
{
	struct ipt_sLog_info *loginfo = (struct ipt_sLog_info *) t->data;

	loginfo->max_num = IPT_SLOG_MAX_ENTRIES;
	loginfo->timeout = IPT_SLOG_TIMEOUT;

	/* Can't cache this */
	*nfcache |= NFC_UNKNOWN;
}


/* Function which parses command options; returns true if it
   ate an option */
static int parse(int c, char **argv, int invert, unsigned int *flags,
		 const struct ipt_entry *entry,
		 struct ipt_entry_target **target)
{
	struct ipt_sLog_info *loginfo =
	    (struct ipt_sLog_info *) (*target)->data;
	int max_num, timeout;

	switch (c) {
	case '1':
		if (*flags & IPT_SLOG_OPT_MAX_NUM)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --sLog-max_num twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --sLog-max_num");
		max_num = atoi(optarg);
		if (max_num > IPT_SLOG_MAX_ENTRIES || max_num < 1)
			exit_error(PARAMETER_PROBLEM,
				   "--sLog-max_num has to be between 1 and '%d'", IPT_SLOG_MAX_ENTRIES);

		loginfo->max_num = max_num;

		*flags |= IPT_SLOG_OPT_MAX_NUM;
		break;

	case '2':
		if (*flags & IPT_SLOG_OPT_TIMEOUT)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --sLog-timeout twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --sLog-timeout");
		
		timeout = atoi(optarg);
		if (timeout > 24*60*60 || timeout < 1)
			exit_error(PARAMETER_PROBLEM,
				   "--sLog-timeout has to be between 1 and '%d'", 24*60*60);
		
		loginfo->timeout = timeout;
		
		*flags |= IPT_SLOG_OPT_TIMEOUT;
		break;

	default:
		return 0;		
	}
	return 1;
}

/* Final check; nothing. */
static void final_check(unsigned int flags)
{
}

/* Saves the union ipt_targinfo in parsable form to stdout. */
static void save(const struct ipt_ip *ip,
		 const struct ipt_entry_target *target)
{
	const struct ipt_sLog_info *loginfo
	    = (const struct ipt_sLog_info *) target->data;

	if ( loginfo )
	{	
		if ( loginfo->max_num != IPT_SLOG_MAX_ENTRIES)
			printf("--sLog-max_num %d ", loginfo->max_num);

		if ( loginfo->timeout != IPT_SLOG_TIMEOUT)
			printf("--sLog-timeout %d ", loginfo->timeout);
	}	
}

/* Prints out the targinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_target *target, int numeric)
{
	const struct ipt_sLog_info *loginfo
	    = (const struct ipt_sLog_info *) target->data;

	if( loginfo )
	{	
		printf("sLog ");
		printf("max_num %d ", loginfo->max_num);
		printf("timeout %d ", loginfo->timeout);
	}	
}

static
struct iptables_target ulog
= {
	.name          = "sLog",
	.version       = IPTABLES_VERSION,
	.size          = IPT_ALIGN(sizeof(struct ipt_sLog_info)),
	.userspacesize = IPT_ALIGN(sizeof(struct ipt_sLog_info)),
	.help          = &help,
	.init          = &init,
	.parse         = &parse,
	.final_check   = &final_check,
	.print         = &print,
	.save          = &save,
	.extra_opts    = opts
};

void _init(void)
{
	register_target(&ulog);
}
