/* dhcpd.c
 *
 * Moreton Bay DHCP Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/time.h>

#include "debug.h"
#include "dhcpd.h"
#include "arpping.h"
#include "socket.h"
#include "options.h"
#include "files.h"
#include "leases.h"
#include "packet.h"
#include "serverpacket.h"
#include "pidfile.h"
#ifdef ACTION_TEC
	#include <dirent.h>

#ifdef ACTION_TEC_QWEST_06
#include "static_leases.h"
#endif
void reloadconfig(char *file);
#ifdef ACTION_TEC_HURL_DEMO
void sethurl(u_int32_t yiaddr);
void getLanIP(char *lanip);
int getHurlFlag();
#endif
#if defined(ACTION_TEC_VLAN) || defined(ACTION_TEC_VIRTUAL_ETH)
#define AMINO_DHCP_SWITCH "/var/etc/dhcp_switch"
int is_option60_found(struct dhcpMessage *oldpacket);
int switch_br01(struct dhcpMessage *oldpacket);
#endif /* defined(ACTION_TEC_VLAN) || defined(ACTION_TEC_VIRTUAL_ETH) */
int gSinglePC = 0;

	#define READ_BUF_SIZE	50


/* find_pid_by_name()
 *  
 *  This finds the pid of the specified process.
 *  Currently, it's implemented by rummaging through 
 *  the proc filesystem.
 *
 */
int exist_pid_by_name( char* pidName,int found_count)
{
	DIR *dir;
	struct dirent *next;
	int i=0;
	int found=0;

	dir = opendir("/proc");
	if (!dir) {
		syslog (LOG_ERR, "can't read /proc");
		return -1;	 //don't do anything
	}
	while ((next = readdir(dir)) != NULL) {
		FILE *status;
		char filename[READ_BUF_SIZE];
		char buffer[READ_BUF_SIZE];
		char name[READ_BUF_SIZE];

		/* Must skip ".." since that is outside /proc */
		if (strcmp(next->d_name, "..") == 0)
			continue;

		/* If it isn't a number, we don't want it */
		if (!isdigit(*next->d_name))
			continue;

		sprintf(filename, "/proc/%s/status", next->d_name);
		if (! (status = fopen(filename, "r")) ) {
			continue;
		}
		if (fgets(buffer, READ_BUF_SIZE-1, status) == NULL) {
			fclose(status);
			continue;
		}
		fclose(status);

		/* Buffer should contain a string like "Name:   binary_name" */
		sscanf(buffer, "%*s %s", name);
		if (strcmp(name, pidName) == 0) {
			found++;
			if(found==found_count){
				closedir(dir);
				return 1;
			}
		}
	}

	closedir(dir);
	return 0;
}

#endif

/* globals */
struct dhcpOfferedAddr *leases;
struct server_config_t server_config[MAX_INTERFACES];
int no_of_ifaces = 0;

#define LEASE_ADD		1
#define LEASE_DEL 	2
extern void write_to_delta(u_int8_t *chaddr, u_int32_t yiaddr, u_int8_t *hname,unsigned long leasetime,u_int8_t action);

/* Exit and cleanup */
static void exit_server(int retval, int ifid)
{
	pidfile_delete(server_config[ifid].pidfile);
	CLOSE_LOG();
	exit(retval);
}


/* SIGTERM handler */
static void udhcpd_killed(int sig, int ifid)
{
	sig = 0;
	LOG(LOG_INFO, "Received SIGTERM");
	exit_server(0, ifid);
}


#ifdef COMBINED_BINARY	
int udhcpd(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{   
	fd_set rfds;
	struct timeval tv;
	int server_socket[MAX_INTERFACES];
	int bytes, retval;
	struct dhcpMessage packet;
	unsigned char *state;
	unsigned char *server_id, *requested;
	u_int32_t server_id_align, requested_align;
	unsigned long timeout_end[MAX_INTERFACES];
	struct option_set *option;
	struct dhcpOfferedAddr *lease;
	int pid_fd;
	int i;
#ifdef ACTION_TEC
	char configfile[80];
#endif

#ifdef CONFIG_ACTIONTEC_TR111
	int ack_ret = 0;
#endif

#if defined(ACTION_TEC_VLAN) || defined(ACTION_TEC_VIRTUAL_ETH)
        int dhcp60_on = 0;
#endif /* defined(ACTION_TEC_VLAN) || defined(ACTION_TEC_VIRTUAL_ETH) */
#ifdef ACTION_TEC
	/*fix bug: Single PC mode, DHCP renew before public IP ready, private for ever */
	FILE *fp;
#ifdef ACTION_TEC_QWEST_06
    //For static IP lease
	struct dhcpOfferedAddr static_lease;
	uint32_t reserved_wanip_lease;
#endif
	fp = fopen("/var/run/udhcpd.pid", "r");
	if (fp) {
		fclose(fp);
		if (exist_pid_by_name ("udhcpd",2)) {
			printf("running dhcpd exsisted, exit!!!\n");
			exit(0);
		} else
			printf("a running dhcpd had a broken-exit!!!\n");
	}
	fp = fopen("/var/tmp/singlepc", "r");
	if (fp) {
		gSinglePC = 1;
		fclose(fp);
	}
#endif
#if defined(ACTION_TEC_VLAN) || defined(ACTION_TEC_VIRTUAL_ETH)
        fp = fopen(AMINO_DHCP_SWITCH, "r");
        if (fp) 
        {
            char sw_buf[256] = {0};
            int p1, p2, p3, p4;
            if (fgets(sw_buf, 255, fp))
            {
                int sw_int = -1;
                if(sw_buf && (strlen(sw_buf) > 0))
                {
                    sw_int = atoi(sw_buf);
                }
                if(sw_int == 1)
                {
                    dhcp60_on = 1;
                    kprintf("#STB# Automatically via DHCP Option 60\n");
                } 
            }
            if(dhcp60_on == 0)
            {
                memset(sw_buf, 0, sizeof(sw_buf));
                if (fgets(sw_buf, 255, fp))
                {
                    sscanf(sw_buf,"%d %d %d %d", &p1, &p2, &p3, &p4);
#if 0                     
                    if(p1 == 1)
                    {
                        kprintf("switch port1 for IPTV\n");
                        system("brctl delif br0 eth0.1");
                        system("echo '0d1' > /proc/sys/dev/marvell/vlan");
                        system("brctl addif br1 eth0.1");
                        system("echo '1a1' > /proc/sys/dev/marvell/vlan");
                    }
                    if(p2 == 1)
                    {
                        kprintf("switch port2 for IPTV\n");
                        system("brctl delif br0 eth0.2");
                        system("echo '0d2' > /proc/sys/dev/marvell/vlan");
                        system("brctl addif br1 eth0.2");
                        system("echo '1a2' > /proc/sys/dev/marvell/vlan");
                    }
                    if(p3 == 1)
                    {
                        kprintf("switch port3 for IPTV\n");
                        system("brctl delif br0 eth0.3");
                        system("echo '0d3' > /proc/sys/dev/marvell/vlan");
                        system("brctl addif br1 eth0.3");
                        system("echo '1a3' > /proc/sys/dev/marvell/vlan");
                    }
                    if(p4 == 1)
                    {
                        kprintf("switch port4 for IPTV\n");
                        system("brctl delif br0 eth0.4");
                        system("echo '0d4' > /proc/sys/dev/marvell/vlan");
                        system("brctl addif br1 eth0.4");
                        system("echo '1a4' > /proc/sys/dev/marvell/vlan");
                    }
#endif
                }
            }
            fclose(fp);
        }
#endif /* defined(ACTION_TEC_VLAN) || defined(ACTION_TEC_VIRTUAL_ETH) */
	OPEN_LOG("udhcpd");
	LOG(LOG_INFO, "udhcp server (v%s) started", VERSION);

	for (i = 0; i < MAX_INTERFACES; i++)
		memset(&server_config[i], 0, sizeof(struct server_config_t));
#ifdef ACTION_TEC
	if (argc < 2) {
		strncpy((char *)&configfile[0],DHCPD_CONF_FILE,sizeof(configfile));
		read_config(DHCPD_CONF_FILE);
	} else {
		strncpy((char *)&configfile[0],argv[1],sizeof(configfile));
		read_config(argv[1]);
	}
#else
	if (argc < 2)
		read_config(DHCPD_CONF_FILE);
	else read_config(argv[1]);
#endif

	if (no_of_ifaces == 0)
		exit(0);

    /*  
	for (i = 0; i < no_of_ifaces; i++)
    totalleases += server_config[i].max_leases;
	leases = malloc(sizeof(struct dhcpOfferedAddr) * totalleases);
	memset(leases, 0, sizeof(struct dhcpOfferedAddr) * totalleases);  */
  
      		
	for (i = 0; i < no_of_ifaces; i++)
	{
		pid_fd = pidfile_acquire(server_config[i].pidfile);
		pidfile_write_release(pid_fd);

		if ((option = find_option(server_config[i].options, DHCP_LEASE_TIME))) {
			memcpy(&server_config[i].lease, option->data + 2, 4);
			server_config[i].lease = ntohl(server_config[i].lease);
		}
		else server_config[i].lease = LEASE_TIME;

		//leases = malloc(sizeof(struct dhcpOfferedAddr) * server_config[i].max_leases);
		//memset(leases, 0, sizeof(struct dhcpOfferedAddr) * server_config[i].max_leases);
    server_config[i].leases = malloc(sizeof(struct dhcpOfferedAddr) * server_config[i].max_leases);
		memset(server_config[i].leases, 0, sizeof(struct dhcpOfferedAddr) * server_config[i].max_leases);
		read_leases(server_config[i].lease_file, i);

		if (read_interface(server_config[i].interface, &server_config[i].ifindex,
						   &server_config[i].server, server_config[i].arp) < 0)
			server_config[i].active = FALSE;
		else
			server_config[i].active	= TRUE;

#ifndef DEBUGGING
		pid_fd = pidfile_acquire(server_config[i].pidfile);	/* hold lock during fork. */
		/* cfgmr req: do not fork */
		/*
		if (daemon(0, 0) == -1) {
			perror("fork");
			exit_server(1, i);
		}
		*/

		pidfile_write_release(pid_fd);
#endif

		signal(SIGUSR1, write_leases);
		signal(SIGTERM, udhcpd_killed);
	}

	for (i = 0; i < no_of_ifaces; i++) {
		timeout_end[i] = time(0) + server_config[i].auto_time;
		server_socket[i] = -1;
		LOG(LOG_INFO, "interface: %s, start : %x end : %x\n", server_config[i].interface, server_config[i].start, server_config[i].end);
	}

	while (1) {	/* loop until universe collapses */
		for (i = 0; i < no_of_ifaces; i++)
		{			
			if (server_config[i].active == FALSE)
				continue;

			if (server_socket[i] < 0)
				if ((server_socket[i] = listen_socket(INADDR_ANY, SERVER_PORT, server_config[i].interface)) < 0) {
					LOG(LOG_ERR, "FATAL: couldn't create server socket, %s", sys_errlist[errno]);
					exit_server(0, i);
				}

			FD_ZERO(&rfds);
			FD_SET(server_socket[i], &rfds);

			if (server_config[i].auto_time) {
				tv.tv_sec = timeout_end[i] - time(0);
				if (tv.tv_sec <= 0) {
					tv.tv_sec = server_config[i].auto_time;
					timeout_end[i] = time(0) + server_config[i].auto_time;
					write_leases(i);
				}
				tv.tv_usec = 0;
			}

#ifdef ACTION_TEC
			tv.tv_sec = 0;
			tv.tv_usec = 400000;
#endif
			retval = select(server_socket[i] + 1, &rfds, NULL, NULL, server_config[i].auto_time ? &tv : NULL);

			if (retval == 0) {
				write_leases(i);
				timeout_end[i] = time(0) + server_config[i].auto_time;
				continue;
			} else if (retval < 0) {
				DEBUG(LOG_INFO, "error on select");
				continue;
			}

			if ((bytes = get_packet(&packet, server_socket[i])) < 0) { /* this waits for a packet - idle */
				if (bytes == -1 && errno != EINTR) {
					DEBUG(LOG_INFO, "error on read, %s, reopening socket", sys_errlist[errno]);
					close(server_socket[i]);
					server_socket[i] = -1;
				}
				continue;
			}

			if ((state = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
				DEBUG(LOG_ERR, "couldn't get option from packet, ignoring");
				continue;
			}
#ifdef ACTION_TEC
			reloadconfig((char *)&configfile[0]);
#endif		

#ifdef ACTION_TEC_QWEST_06
              /* For static IP lease */
              /* Look for a static lease */
            if (strcmp(server_config[i].actiontec_vip_enable, "0") != 0)
                reserved_wanip_lease = getIpByMac(server_config[i].reserved_WanIp,
                        &packet.chaddr);
            else
                reserved_wanip_lease = NULL;

            if(reserved_wanip_lease)
            {
                memcpy(&static_lease.chaddr, &packet.chaddr, 16);
                static_lease.yiaddr = reserved_wanip_lease;
                static_lease.expires = 0;
                lease = &static_lease;
		
            } /* For static IP lease end */
            else
#endif
                  /* ADDME: look for a static lease */
                lease = find_lease_by_chaddr(packet.chaddr, i);
#if defined(ACTION_TEC_VLAN) || defined(ACTION_TEC_VIRTUAL_ETH)
                        if (dhcp60_on == 1)
                        {
                        if ((state[0] == DHCPDISCOVER) || (state[0] == DHCPREQUEST) || (state[0] == DHCPDECLINE) || (state[0] == DHCPRELEASE) || (state[0] == DHCPINFORM))
                        {
                            int is_op60 = is_option60_found(&packet);
                            if((is_op60 == 1) && (i == 0))
                            {
                                switch_br01(&packet); 
                                continue;
                            } else if (i == 1)
                            {
                                if (is_op60 == 0)
                                {
                                    ;
                                }
                                continue;
                            }
                        }
                        }
#endif /* defined(ACTION_TEC_VLAN) || defined(ACTION_TEC_VIRTUAL_ETH) */
			switch (state[0]) {
			case DHCPDISCOVER:
                                DEBUG(LOG_INFO,"received DISCOVER");
                                        if (sendOffer(&packet, i) < 0)
                                        LOG(LOG_INFO, "send OFFER failed");
				break;          

			case DHCPREQUEST:
				DEBUG(LOG_INFO, "received REQUEST");
				requested = get_option(&packet, DHCP_REQUESTED_IP);
				server_id = get_option(&packet, DHCP_SERVER_ID);
				if (requested) memcpy(&requested_align, requested, 4);
				if (server_id) memcpy(&server_id_align, server_id, 4);

					if (lease)
          { /*ADDME: or static lease */
						if (server_id)
            {
						/* SELECTING State */
						DEBUG(LOG_INFO, "server_id = %08x", ntohl(server_id_align));
						if (server_id_align == server_config[i].server && requested && 
					    		requested_align == lease->yiaddr)
#ifdef ACTION_TEC_QWEST_06
                        {
#endif //defined ACTION_TEC_QWEST_06
					    	#ifdef CONFIG_ACTIONTEC_TR111
					    	{
					    		ack_ret = 
						#endif
						
							sendACK(&packet, lease->yiaddr, i);

						#ifdef CONFIG_ACTIONTEC_TR111
							if (ack_ret >= 0)
								deal_vendor_data(&packet, lease->yiaddr, DHCP_VENDOR_SPECIFIC_INFO, ACTION_ADD_VENDOR_INFO);
					    	}
						#endif
#ifdef ACTION_TEC_HURL_DEMO
sethurl(lease->yiaddr);
#endif
						}
#ifdef ACTION_TEC_QWEST_06
                        else
                            sendNAK(&packet, i);
            }
#endif //defined ACTION_TEC_QWEST_06
            else
            {
							if (requested)
              {
							/* INIT-REBOOT State */
							if (lease->yiaddr == requested_align){
								#ifdef CONFIG_ACTIONTEC_TR111
						    		{
						    			ack_ret =
								#endif
							
								sendACK(&packet, lease->yiaddr, i);
								
								#ifdef CONFIG_ACTIONTEC_TR111
									if (ack_ret >= 0)
											deal_vendor_data(&packet, lease->yiaddr, DHCP_VENDOR_SPECIFIC_INFO, ACTION_ADD_VENDOR_INFO);
								}
								#endif
#ifdef ACTION_TEC_HURL_DEMO
sethurl(lease->yiaddr);
#endif
								}
								else
                  sendNAK(&packet, i);
							}
              else
              {
							/* RENEWING or REBINDING State */
							if (lease->yiaddr == packet.ciaddr){
								#ifdef CONFIG_ACTIONTEC_TR111
									ack_ret =
								#endif
								
								sendACK(&packet, lease->yiaddr, i);
		
								#ifdef CONFIG_ACTIONTEC_TR111
									if (ack_ret >= 0)
										deal_vendor_data(&packet, lease->yiaddr, DHCP_VENDOR_SPECIFIC_INFO, ACTION_ADD_VENDOR_INFO);
								#endif

#ifdef ACTION_TEC_HURL_DEMO
sethurl(lease->yiaddr);
#endif
								}
								else
								/* don't know what to do!!!! */
								sendNAK(&packet, i);
							}                       
						}
					/* what to do if we have no record of the client */
					}
          else if (server_id)
					/* SELECTING State */
					sendNAK(&packet,i);
          else if (requested)
          {
					/* INIT-REBOOT State */
					  if ((lease = find_lease_by_yiaddr(requested_align,i)))
            {
						if (lease_expired(lease,i))
							/* probably best if we drop this lease */
							memset(lease->chaddr, 0, 16);
						/* make some contention for this address */
						else
							sendNAK(&packet,i);
			      }
            else if (requested_align < server_config[i].start ||
							   requested_align > server_config[i].end)
						sendNAK(&packet,i);
					else
						/* else remain silent */
						sendNAK(&packet,i);
		      }
          else
					/* RENEWING or REBINDING State */
					sendNAK(&packet,i);
				break;
			case DHCPDECLINE:
				DEBUG(LOG_INFO,"received DECLINE");
				if (lease) {
					memset(lease->chaddr, 0, 16);
					lease->expires = time(0) + server_config[i].decline_time;
				}
				break;

			case DHCPRELEASE:
				DEBUG(LOG_INFO,"received RELEASE");
					if (lease)
					{
						if( lease->expires != server_config[i].inflease_time)
						{
						lease->expires = time(0);
						write_to_delta(lease->chaddr, lease->yiaddr, lease->hostname,0,LEASE_DEL);
					}
				}
				break;

			case DHCPINFORM:
				DEBUG(LOG_INFO,"received INFORM");
				send_inform(&packet, i);
				break;  

			default:
				LOG(LOG_WARNING, "unsupported DHCP message (%02x) -- ignoring", state[0]);
			}
		}
	}
	return 0;
}


#ifdef CONFIG_ACTIONTEC_TR111
//action == 0 : insert vendor information into file
//action == 1 : remove vendor information from file
void deal_vendor_data(struct dhcpMessage *packet, u_int32_t yiaddr, int code, int action)
{
	unsigned char *point_vendor = NULL;
	unsigned char *vendor_info = NULL;
	int opt_len = 0;
	int pos = 0, sub_pos = 0;
	unsigned char enterprise_no[5];
	u_int8_t option_data_len = 0;
	unsigned char option_data[160];
	unsigned char manufacturer_OUI[10];
	unsigned char serial_number[70];
	unsigned char product_class[70];
	u_int8_t sub_option_code;
	u_int8_t sub_option_len;
	unsigned char option_flag = 0x00;

	point_vendor = get_option(packet, code);
	if (point_vendor){
		opt_len = get_option_length(packet, code);
		vendor_info = (unsigned char *)malloc(opt_len + 1);
	}
	if (vendor_info && point_vendor){
		memset(vendor_info, 0, opt_len + 1);
		memcpy(vendor_info, point_vendor, opt_len);
		memset(manufacturer_OUI, 0, sizeof(manufacturer_OUI));
		memset(serial_number, 0, sizeof(serial_number));
		memset(product_class, 0, sizeof(product_class));
		while (pos < opt_len){//get more than one enterprise information
			memset(option_data, 0, sizeof(option_data));
			memset(manufacturer_OUI, 0, sizeof(manufacturer_OUI));
			memset(serial_number, 0, sizeof(serial_number));
			memset(product_class, 0, sizeof(product_class));
			memset(enterprise_no, 0, sizeof(enterprise_no));
			sub_pos = 0;
			memcpy(enterprise_no, vendor_info + pos, sizeof(enterprise_no) - 1);

			if (strncmp(enterprise_no, DSLFORUM_ENTERPRISE_NO, strlen(enterprise_no)) != 0)
				return;
			else	{// get one enterprise information
				option_data_len = vendor_info[pos + ENTERPRISE_NO_LENGTH];
				memcpy(option_data, &vendor_info[pos + ENTERPRISE_NO_LENGTH + OPTION_DATA_LENGTH], option_data_len);
				while (sub_pos < option_data_len){
					sub_option_code = option_data[sub_pos];
					sub_option_len = option_data[sub_pos + SUB_OPTION_CODE_LENGTH];
					if (sub_option_code == DEVICE_MANUFACTURER_OUI){
						memcpy(manufacturer_OUI, &option_data[sub_pos + SUB_OPTION_CODE_LENGTH + SUB_OPTION_DATA_LENGTH], sub_option_len);
						option_flag |= 0x01;
					}
					else if (sub_option_code == DEVICE_SERIAL_NUMBER){
						memcpy(serial_number, &option_data[sub_pos + SUB_OPTION_CODE_LENGTH + SUB_OPTION_DATA_LENGTH], sub_option_len);
						option_flag |= 0x02;
					}
					else if (sub_option_code == DEVICE_PRODUCT_CLASS){
						memcpy(product_class, &option_data[sub_pos + SUB_OPTION_CODE_LENGTH + SUB_OPTION_DATA_LENGTH], sub_option_len);
						option_flag |= 0x04;
					}
					sub_pos += SUB_OPTION_CODE_LENGTH + SUB_OPTION_DATA_LENGTH + sub_option_len;
				}
				//have get all device information, so write these information into file
				if (option_flag == 0x07){
					if (action == ACTION_ADD_VENDOR_INFO){
						insert_into_file(yiaddr, manufacturer_OUI, serial_number, product_class);
						LOG(LOG_INFO, "add new vendor: %s %s, %s", (char*)manufacturer_OUI, serial_number, product_class);
					}
					else if (action == ACTION_REMOVE_VENDOR_INFO){
						delete_from_file(yiaddr, manufacturer_OUI, serial_number, product_class);
						LOG(LOG_INFO, "delete old vendor: %s %s, %s", (char*)manufacturer_OUI, serial_number, product_class);
					}
					option_flag &= 0x00;
				}
			}
			pos += ENTERPRISE_NO_LENGTH + OPTION_DATA_LENGTH + option_data_len;
		}// end while (pos < opt_len)
	}
}
#endif

#ifdef CONFIG_ACTIONTEC_TR111
void delete_from_file(u_int32_t yiaddr, unsigned char * oui, unsigned char * sno, unsigned char * pclass)
{
	char info_filename[] = "/var/tmp/deviceinfo_tr111.txt";
	char temp_filename[] = "/var/tmp/deviceinfo_tr111.txt.tmp";
	char vendor_info[256];
	char buffer[256];
	FILE *fp_old = NULL;
	FILE *fp_new = NULL;

	if ((fp_old=fopen(info_filename, "r")) == NULL)
		return;

	if ((fp_new=fopen(temp_filename, "w")) == NULL){
		fclose(fp_old);
		return;
	}

	sprintf(vendor_info, "%u %s %s %s\n", yiaddr, (char*)oui, (char*)sno, (char*)pclass);

	while(fgets(buffer, sizeof(buffer), fp_old)){
		if (strncmp(buffer, vendor_info, strlen(vendor_info)) == 0)
			continue;
		fprintf(fp_new, "%s", buffer);
	}
	fclose(fp_old);
	fclose(fp_new);
	unlink(info_filename);
	rename(temp_filename, info_filename);
}
#endif

#ifdef CONFIG_ACTIONTEC_TR111
void insert_into_file(u_int32_t yiaddr, unsigned char * oui, unsigned char * sno, unsigned char * pclass)
{
	char info_filename[] = "/var/tmp/deviceinfo_tr111.txt";
	FILE *fp = NULL;
	if (oui == NULL || sno == NULL || pclass == NULL)
		return;

	if (find_record_in_file(info_filename, "%u %s %s %s", yiaddr, (char*)oui, (char*)sno, (char*)pclass) == TRUE)
		return;
	
	if ((fp=fopen(info_filename, "a+")) != NULL){
		fprintf(fp, "%u %s %s %s\n", yiaddr, (char*)oui, (char*)sno, (char*)pclass);
		fclose(fp);
	}
		
}
#endif

#ifdef CONFIG_ACTIONTEC_TR111
int find_record_in_file(char * filename, char * fmt,...)
{
	FILE *fp = NULL;
	va_list ap;
	char buff[256];
	char file_data[256];
	
	if (filename != NULL){
		if ((fp=fopen(filename, "r")) != NULL){
			va_start(ap, fmt);
			vsnprintf(buff, sizeof(buff) - 1, fmt, ap);
			va_end(ap);
			strcat(buff, "\n");
			while (fgets(file_data, sizeof(file_data), fp)){
				if (strncmp(file_data, buff, strlen(buff)) == 0){
					fclose(fp);
					return TRUE;
				}
			}
			fclose(fp);
		}	
	}
	return FALSE;
}

#endif

#ifdef ACTION_TEC
void reloadconfig(char *file)
{
	struct stat statbuf;
	time_t resolv_changed = -1;

	if ((stat(file, &statbuf) == 0) && 
		(statbuf.st_mtime > resolv_changed) &&
		(statbuf.st_mtime < time(NULL) || resolv_changed == -1)) {
#ifdef ACTION_TEC //DEBUG
		fprintf(stderr,"Load new config %s\n",file);
#endif      
		resolv_changed = statbuf.st_mtime;
		read_config(file);
	}
}
#ifdef ACTION_TEC_HURL_DEMO
void sethurl(u_int32_t yiaddr)
{ char *file="/var/tmp/dhcp_iphost_hurl";
  FILE *fp=NULL;
  char tmp[100];
  char state[3];
  char cmd[200];
  struct in_addr addr;
  char *ip;
  char lanip[20]={0}; 
  
  if (getHurlFlag()!=1) {
      return;
  }
  addr.s_addr =yiaddr;

  ip=inet_ntoa(addr);

  fp=fopen(file,"a+");
  //kprintf("***HURLCMD: ip=%s \n",ip);

  if (!fp) {
      //kprintf("open file error\n");
      return;
  }

  rewind(fp);

  while(!feof(fp)){
     memset(tmp,0,100);
     fscanf(fp,"%s %s\n",tmp,state);
     if ((tmp[0]!='\0') && (strncmp(tmp,ip,strlen(ip))==0) ) {
         fclose(fp);
         return;
     }
  }
  fprintf(fp,"%s %s\n", ip, "1");
  fclose(fp); 
  memset(cmd,0,200);
  getLanIP(lanip);
  if (lanip[0]!='\0') {  
      sprintf(&cmd[0],"/var/etc/hurl br0 %s add 0 %s &", lanip,ip);
      system(&cmd[0]);
      //kprintf("***HURLCMD: %s \n",cmd);
  }
}

void getLanIP(char *lanip){
struct in_addr s_ip, s_mask, s_target;
    FILE *fp1;
    char buf[500], ipbuf[100];
    char lan_ip[20] = {0};
    char lan_ipmask[20] = {0};
    
       memset(ipbuf, 0, sizeof(ipbuf));
       memset(buf, 0, sizeof(buf));
       fp1 = fopen("/var/etc/pc/lan_settings", "r");
       if (fp1) {
           if (fgets(buf, sizeof(buf) -1 , fp1) != NULL) {
               char *s, *s1;
               s = strtok (buf," \t\n");
               strcpy (ipbuf, s);
               s1 = strstr(ipbuf, "/");
               if(s1)
               {
                   strncpy(lan_ip, ipbuf, s1-ipbuf);
                   strcpy(lan_ipmask, s1+1);
                   strcpy(lanip,lan_ip);
               }
           }
           fclose(fp1);           
       }       
}

int getHurlFlag(){
    FILE *fp=NULL;
    char buf[10]={0};
    int ret=0;

    fp=fopen("/var/tmp/enable_hurl","r");
    if (fp) {
        if (fgets(buf, sizeof(buf) -1 , fp) != NULL) {
            ret=atoi(buf);
        }        
        fclose(fp);
    }    
       
    return ret;
}
#endif
#endif
#if defined(ACTION_TEC_VLAN) || defined(ACTION_TEC_VIRTUAL_ETH)
#define AMINO_OPTION60_STRING "/var/etc/op60_string"
int is_option60_found(struct dhcpMessage *oldpacket)
{
    int length = 0;
    char option60_name[32] = {0};
    char *opname;
    char buf[512] = {0};
    FILE* fp_op60;
    int ret = 0;
    opname = get_option(oldpacket, DHCP_VENDOR);
    if(opname)
    {
        length = get_option_length(oldpacket, DHCP_VENDOR);
        if (length > 31)
            length = 31;
        strncpy(option60_name, opname, length);
    }
    if((option60_name) && (strlen(option60_name) > 0))
    {
        fp_op60 = fopen(AMINO_OPTION60_STRING, "r");
        if(fp_op60)
        {
            if (fgets(buf, 511, fp_op60))
            {
                char *s;
                fclose(fp_op60);
                s = strtok(buf, " \r\n");
                while(s)
                {
                    if(s && (strlen(s) > 0))
                    {
                        //kprintf("s=%s\n", s);
                        if(strncmp(option60_name, s, strlen(s)) == 0)
                        {
                            kprintf("#STB# Option 60: %s\n", option60_name);
                            ret = 1;
                            break;
                        }
                    }
                    s = strtok(NULL, " \r\n");
                }
            } else {
                fclose(fp_op60);
            }
        }
    }
    return ret;
}
int switch_br01(struct dhcpMessage *oldpacket)
{
    char option60_mac[18] = {0};
    char tmp_port[100] = {0}, tmp_name[100] = {0}, tmp_mac[100] = {0}, tmp_local[100] = {0}, tmp_time[100] = {0};
    FILE* fp2;
    int eth_id = 0;
    sprintf(option60_mac, "%02x:%02x:%02x:%02x:%02x:%02x",
        oldpacket->chaddr[0], oldpacket->chaddr[1], oldpacket->chaddr[2], oldpacket->chaddr[3], oldpacket->chaddr[4], oldpacket->chaddr[5]);
    system("brctl showifs br0 > /var/tmp/port_mac &");
    sleep(1);
    fp2 = fopen("/var/tmp/port_mac", "r");
    if(fp2)
    {
        while(!feof(fp2))
        {
            memset(tmp_port, 0, sizeof(tmp_port));
            memset(tmp_name, 0, sizeof(tmp_name));
            memset(tmp_mac, 0, sizeof(tmp_mac));
            memset(tmp_local, 0, sizeof(tmp_local));
            memset(tmp_time, 0, sizeof(tmp_time));
            fscanf(fp2,"%s\t%s\t%s\t%s\n", tmp_port, tmp_name, tmp_mac, tmp_local, tmp_time);
            if ((tmp_mac) && (tmp_mac[0] != '\0') && (strlen(tmp_mac) > 0))
            {
                if (strncmp(tmp_mac, option60_mac, strlen(option60_mac)) == 0)
                {
                    kprintf("#STB# Found mac:%s on interface:%s\n", tmp_mac, tmp_name);
                    if (tmp_name)
                    {
                        char *s1;
                        s1 = strstr(tmp_name, "eth0.");
                        if (s1)
                        {
                            s1 = s1 + strlen("eth0.");
                            if (s1)
                            {
                                eth_id = atoi(s1);
                                kprintf("#STB# port %d\n", eth_id);
                            }
                        }
                    } 
                    if((eth_id > 0) && (eth_id < 5))
                    {
                        char cmd_buf[256] = {0};
                        char eth_str[10] = {0};
                        FILE *fp3;
                        int pval[4] = {-1, -1, -1, -1};
                        /*eth_id = eth_id - 1;*/
                        fp3 = fopen(AMINO_DHCP_SWITCH, "r");
                        if (fp3)
                        {
                            if(fgets(cmd_buf, 255, fp3))
                            {
                                 memset(cmd_buf, 0, sizeof(cmd_buf));
                                 if (fgets(cmd_buf, 255, fp3))
                                 {
                                     sscanf(cmd_buf,"%d %d %d %d", &pval[0], &pval[1], &pval[2], &pval[3]);
                                 }
                            }
                            fclose(fp3);
                        }
#if 0
#ifndef ACTION_TEC_VIRTUAL_ETH
                        sprintf(eth_str, "eth0.%d", eth_id);
#else /* ACTION_TEC_VIRTUAL_ETH */
                        sprintf(eth_str, "cp.%d", eth_id);
#endif /* ACTION_TEC_VIRTUAL_ETH */
                        sprintf(cmd_buf, "brctl delif br0 %s", eth_str);
                        system(cmd_buf);
                        memset(cmd_buf, 0, sizeof(cmd_buf));
			sprintf(cmd_buf, "echo '0d%d' > /proc/sys/dev/marvell/vlan", eth_id);
                        system(cmd_buf);
                        memset(cmd_buf, 0, sizeof(cmd_buf));
                        sprintf(cmd_buf, "brctl addif br1 %s", eth_str);
                        system(cmd_buf);
                        memset(cmd_buf, 0, sizeof(cmd_buf));
                        sprintf(cmd_buf, "echo '1a%d' > /proc/sys/dev/marvell/vlan", eth_id);
                        system(cmd_buf);
#endif
                        if(pval[eth_id - 1] == 0)
                        {
                            memset(cmd_buf, 0, sizeof(cmd_buf));
                            sprintf(cmd_buf, "single_settings.sh amino port%d %d", eth_id, 1);
                            system(cmd_buf);
                        }
                        break;
                    }
                }
            }
        }
        fclose(fp2);
        unlink("/var/tmp/port_mac");
    }
    return eth_id;
}
#endif /* defined(ACTION_TEC_VLAN) || defined(ACTION_TEC_VIRTUAL_ETH) */

